'''
Storage for common doc strings and templates shared in non-related classes and methods.
'''


import typing as tp

from static_frame.core.util import AnyCallable

#-------------------------------------------------------------------------------
# common strings
#NOTE: for kwargs, it is sometimes useful to only define the string, not the variable name, as in some contexts different variable names are use same conceptual entity.
INDEX_INITIALIZER = '''An iterable of unique, hashable values, or another ``Index`` or ``IndexHierarchy``, to be used as the labels of the index.'''

LOC_SELECTOR = '''A loc selector, either a label, a list of labels, a slice of labels, or a Boolean array.'''

ILOC_SELECTOR = '''An iloc selector, either an index, a list of indices, a slice of indices, or a Boolean array.'''


#-------------------------------------------------------------------------------
# full parameter definitions

AXIS = '''axis: Integer specifying axis, where 0 is rows and 1 is columns. Axis 0 is set by default.'''

CHUNKSIZE = 'chunksize: Units of work per executor, as passed to the Thread- or ProcessPoolExecutor.'

COLUMNS_CONSTRUCTOR = "columns_constructor: Optional class or constructor function to create the :obj:`Index` applied to the columns."

CONSOLIDATE_BLOCKS = 'consolidate_blocks: Optionally consolidate adjacent same-typed columns into contiguous arrays.'

DEEPCOPY_FROM_BUS = 'deepcopy_from_bus: Boolean to determine if containers are deep-copied from the contained :obj:`Bus` during extraction. Set to ``True`` to avoid holding references from the :obj:`Bus`.'

DTYPE_SPECIFIER = '''dtype: A value suitable for specyfying a NumPy dtype, such as a Python type (float), NumPy array protocol strings ('f8'), or a dtype instance.'''

DTYPES = "dtypes: Optionally provide an iterable of dtypes, equal in length to the length of each row, or a mapping by column name. If a dtype is given as None, NumPy's default type determination will be used."

FP = 'fp: A string file path or :obj:`Path` instance.'

INDEX_CONSTRUCTOR = "index_constructor: Optional class or constructor function to create the :obj:`Index` applied to the rows."

MAX_PERSIST = 'max_persist: When loading :obj:`Frame` from a :obj:`Store`, optionally define the maximum number of :obj:`Frame` to remain in the :obj:`Bus`, regardless of the size of the :obj:`Bus`. If more than ``max_persist`` number of :obj:`Frame` are loaded, least-recently loaded :obj:`Frame` will be replaced by ``FrameDeferred``. A ``max_persist`` of 1, for example, permits reading one :obj:`Frame` at a time without ever holding in memory more than 1 :obj:`Frame`.'

MAX_WORKERS = 'max_workers: Number of parallel executors, as passed to the Thread- or ProcessPoolExecutor; ``None`` defaults to the max number of machine processes.'

NAME = 'name: A hashable object to label the container.'

OWN_COLUMNS = '''own_columns: Flag the passed columns as ownable by this :obj:`static_frame.{class_name}`. Primarily used by internal clients.'''

OWN_DATA = '''own_data: Flag the data values as ownable by this :obj:`static_frame.{class_name}`. Primarily used by internal clients.'''

OWN_INDEX = '''own_index: Flag the passed index as ownable by this :obj:`static_frame.{class_name}`. Primarily used by internal clients.'''

RETAIN_LABELS = 'retain_labels: Boolean to determine if, along the axis of virtual concatentation, if component :obj:`Frame` labels should be used to form the outer depth of an :obj:`IndexHierarchy`. This is required to be ``True`` if component :obj:`Frame` labels are not globally unique along the axis of concatenation.'

STORE = 'store: A :obj:`Store` subclass.'

STORE_CONFIG_MAP = 'config: A :obj:`StoreConfig`, or a mapping of label ot :obj:`StoreConfig`'

USE_THREADS = 'use_threads: Use the ThreadPoolExecutor instead of the ProcessPoolExecutor.'

class DOC_TEMPLATE:

    #---------------------------------------------------------------------------
    # complete or partial function doc strings

    to_html = '''
    Return an HTML table representation of this :obj:`{class_name}` using standard TABLE, TR, and TD tags. This is not a complete HTML page.

    Args:
        config: Optional :obj:`DisplayConfig` instance.

    Returns:
        :obj:`str`
    '''

    to_html_datatables = '''
    Return a complete HTML representation of this :obj:`{class_name}` using the DataTables JS library for table naviagation and search. The page links to CDNs for JS resources, and thus will not fully render without an internet connection.

    Args:
        fp: optional file path to write; if not provided, a temporary file will be created. Note: the caller is responsible for deleting this file.
        show: if True, the file will be opened with a webbrowser.
        config: Optional :obj:`DisplayConfig` instance.

    Returns:
        :obj:`str`, absolute file path to the file written.
    '''

    name = '''
    A hashable label attached to this container.

    Returns:
        :obj:`Hashable`
    '''

    interface = '''
    A :obj:`Frame` documenting the interface of this class.
    '''

    label_widths_at_depth = '''
    A generator of pairs, where each pair is the label and the count of that label found at the depth specified by  ``depth_level``.

    Args:
        depth_level: a depth level, starting from zero.
    '''

    clip = '''Apply a clip operation to this :obj:`{class_name}`. Note that clip operations can be applied to object types, but cannot be applied to non-numerical objects (e.g., strings, None)'''


    values_2d = '''
    A 2D NumPy array of all values in the :obj:`{class_name}`. As this is a single array, heterogenous columnar types might be coerced to a compatible type.
    '''

    values_1d = '''
    A 1D NumPy array of the values in the :obj:`{class_name}`. This array will have the same dtype as the container.
    '''


    #---------------------------------------------------------------------------
    # dict entries

    apply = dict(
            doc='Apply a function to each value.',
            func='func: A function that takes a value.',
            name=NAME,
            dtype=DTYPE_SPECIFIER,
            max_workers=MAX_WORKERS,
            chunksize=CHUNKSIZE,
            use_threads=USE_THREADS,
            )

    argminmax = dict(
            skipna='skipna: if True, NaN or None values will be ignored; if False, a found NaN will propagate.',
            axis='axis: Axis upon which to evaluate contiguous missing values, where 0 is vertically (between row values) and 1 is horizontally (between column values).'
    )

    astype = dict(
            dtype=DTYPE_SPECIFIER
            )


    batch_constructor = dict(
            args = f'''
        Args:
            {FP}
            {STORE_CONFIG_MAP}
            {MAX_WORKERS}
            {CHUNKSIZE}
            {USE_THREADS}
            '''
            )

    batch_init = dict(
            args = f'''
        Args:
            {NAME}
            {STORE_CONFIG_MAP}
            {MAX_WORKERS}
            {CHUNKSIZE}
            {USE_THREADS}
            '''
            )

    bus_constructor = dict(
            args = f'''
        Args:
            {FP}
            {STORE_CONFIG_MAP}
            {MAX_PERSIST}
            '''
            )

    bus_init = dict(
            args = f'''
        Args:
            {STORE}
            {STORE_CONFIG_MAP}
            {MAX_PERSIST}
            '''
            )

    container_init = dict(
            index='''index: Optional index initializer. If provided in addition to data values, lengths must be compatible.''',
            columns='''columns: Optional column initializer. If provided in addition to data values, lengths must be compatible.
            ''',
            own_index=OWN_INDEX,
            own_data=OWN_DATA,
            own_columns=OWN_COLUMNS
            )

    constructor_frame = dict(
            dtypes=DTYPES,
            name=NAME,
            consolidate_blocks=CONSOLIDATE_BLOCKS
    )
    delimited = dict(
            doc='Given a file path or file-like object, write the :obj:`Frame` as delimited text.',
            fp='A file path, PathLib instance, or file-like object.',
            delimiter='delimiter: Character to be used for delimiterarating elements.',
            include_index='include_index: If True, the index will be written.',
            include_index_name='include_index_name: If including columns, populate the row above the index with the index ``name``. Cannot be True if ``include_columns_name`` is ``True``.',
            include_columns='include_columns: If ``True``, the columns will be written.',
            include_columns_name='include_columns_name: If including index, populate the column to the left of the columns with the columns ``name``. Cannot be True if ``include_index_name`` is True.',
            encoding='encoding: Encoding type to be used when opening the file.',
            line_terminator='line_terminator: The string used to terminate lines.',
            quote_char='quote_char: A one-character string used to quote fields containing special characters, such as the ``delimiter`` or ``quote_char``, or which contain new-line characters.',
            quote_double='quote_double: Controls how instances of quote_char appearing inside a field should themselves be quoted. When ``True``, the character is doubled. When ``False``, the ``escape_char`` is used as a prefix to the ``quote_char``. It defaults to True.',
            escape_char='escape_char: A one-character string used by the writer to escape the delimiter if quoting is set to QUOTE_NONE and the quotechar if quote_double is False.',
            quoting='quoting: Controls when quotes should be generated. It can take on any of the QUOTE_* constants from the standard library csv module.',
            store_filter='store_filter: A :obj:`StoreFilter` instance.',
            )
    duplicated = dict(
            exclude_first='exclude_first: Boolean to select if the first duplicated value is excluded.',
            exclude_last='exclude_last: Boolean to select if the last duplicated value is excluded.',
            axis=AXIS,
            )

    display = dict(
            doc='Return a :obj:`static_frame.Display`, capable of providing a string representation.',
            config='config: A :obj:`static_frame.DisplayConfig` instance. If not provided, the :obj:`static_frame.DisplayActive` will be used.'
            )

    equals = dict(
            doc='Return a :obj:`bool` from comparison to any other object.',
            compare_name="compare_name: Include equality of the container's name (and all composed containers) in the comparison.",
            compare_dtype="compare_dtype: Include equality of the container's dtype (and all composed containers) in the comparison.",
            compare_class="compare_class: Include equality of the container's class (and all composed containers) in the comparison.",
            skipna="skipna: If True, comparisons between missing values are equal.",
            )

    from_any = dict(
            fp=FP,
            index_depth='index_depth: integer specification of how many columns to use in forming the index. A value of 0 will select none; a value greater than 1 will create an :obj:`IndexHierarchy`.',
            columns_depth='columns_depth: integer specification of how many rows to use in forming the columns. A value of 0 will select none; a value greater than 1 will create an :obj:`IndexHierarchy`.',
            columns_select='columns_select: An optional iterable of column names to load.',
            consolidate_blocks=CONSOLIDATE_BLOCKS,
            name=NAME,
            dtypes=DTYPES,
            )


    from_pandas = dict(
            own_data='''own_data: If True, the underlying NumPy data array will be made immutable and used without a copy.''',
            own_index='''own_index: If True, the underlying NumPy index label array will be made immutable and used without a copy.''',
            own_columns='''own_columns: If True, the underlying NumPy column label array will be made immutable and used without a copy.''',
            columns_constructor=COLUMNS_CONSTRUCTOR,
            index_constructor=INDEX_CONSTRUCTOR,
            consolidate_blocks=CONSOLIDATE_BLOCKS,
            )


    fillna = dict(
            limit='limit: Set the maximum count of missing values (NaN or None) to be filled per contiguous region of missing vlaues. A value of 0 is equivalent to no limit.',
            value='value: Value to be used to replace missing values (NaN or None).',
            axis='axis: Axis upon which to evaluate contiguous missing values, where 0 is vertically (between row values) and 1 is horizontally (between column values).'
            )

    head = dict(
            doc='''Return a :obj:`{class_name}` consisting only of the top elements as specified by ``count``.
            ''',
            count='''count: Number of elements to be returned from the top of the :obj:`{class_name}`''',
            )

    index_init = dict(
            args = f'''
        Args:
            labels: {INDEX_INITIALIZER}
            {NAME}
            loc_is_iloc: Optimization when a contiguous integer index is provided as labels. Generally only set by internal clients.
            {DTYPE_SPECIFIER}'''
            )

    index_date_time_init = dict(
            args = f'''
        Args:
            labels: Iterable of hashable values to be used as the index labels. If strings, NumPy datetime conversions will be applied.
            {NAME}
            '''
            )

    insert = dict(
            key_before="key: Label before which the new container will be inserted.",
            key_after="key: Label after which the new container will be inserted.",
            container="container: Container to be inserted.",
            fill_value='fill_value: A value to be used to fill space after reindexing the new container.'
            )

    join = dict(
            left_depth_level="left_depth_level: Specify one or more left index depths to include in the join predicate.",
            left_columns="left_columns: Specify one or more left columns to include in the join predicate.",
            right_depth_level="right_depth_level: Specify one or more right index depths to include in the join predicate.",
            right_columns="right_columns: Specify one or more right columns to include in the join predicate.",
            left_template="left_template: Provide a format string for naming left columns in the joined result.",
            right_template="right_template: Provide a format string for naming right columns in the joined result.",
            fill_value='fill_value: A value to be used to fill space created in the join.',
            composite_index='composite_index: If True, an index of tuples will be returned, formed from the left index label and the right index label; if False, an index of matching labels, if unique, will be returned.',
            composite_index_fill_value='composite_index_fill_value: Value to be used when forming a composite index when a label is missing.'
            )

    mloc = dict(
            doc_int='The memory location, represented as an integer, of the underlying NumPy array.',
            doc_array='The memory locations, represented as an array of integers, of the underlying NumPy arrays.',
            )

    map_any = dict(
            doc='Apply a mapping; for values not in the mapping, the value is returned.',
            mapping='mapping: A mapping type, such as a dictionary or Series.',
            dtype=DTYPE_SPECIFIER,
            )

    map_fill = dict(
            doc = 'Apply a mapping; for values not in the mapping, the ``fill_value`` is returned.',
            mapping = 'mapping: A mapping type, such as a dictionary or Series.',
            fill_value = 'fill_value: Value to be returned if the values is not a key in the mapping.',
            dtype=DTYPE_SPECIFIER
            )

    map_all = dict(
            doc = 'Apply a mapping; for values not in the mapping, an Exception is raised.',
            mapping='mapping: A mapping type, such as a dictionary or Series.',
            dtype=DTYPE_SPECIFIER
            )


    quilt_constructor = dict(
            args = f'''
        Args:
            {FP}
            {STORE_CONFIG_MAP}
            axis: Integer specifying axis of virtual concatenation, where 0 is vertically (stacking rows) and 1 is horizontally (extending columns).
            {RETAIN_LABELS}
            {DEEPCOPY_FROM_BUS}
            {MAX_PERSIST}
            '''
            )

    quilt_init = dict(
            args = f'''
        Args:
            bus: :obj:`Bus` of :obj:`Frame` to be used for virtual concatenation.
            axis: Integer specifying axis of virtual concatenation, where 0 is vertically (stacking rows) and 1 is horizontally (extending columns).
            {RETAIN_LABELS}
            {DEEPCOPY_FROM_BUS}
            '''
            )
    rank = dict(
            skipna='''skipna: If ``True``, exclude NA values (NaN or None) from ranking, replacing those values with ``fill_value``. ''',
            ascending='''ascending: If ``True``, the lowest ranks correspond to the lowest values. The default is ``True``.''',
            ascendings='''ascending: Boolean, or iterable of Booleans; if ``True``, the lowest ranks correspond to the lowest values; if an iterable, apply per column or row. The default is ``True``.''',
            start='''start: The reference value for the lowest rank. Some ranking methodologies (mean, max) may not return this value given some inputs. The default is 0; for ranks that start from 1, provide a value of 1.''',
            fill_value='''fill_value: A value to be used to fill NA values ignored in ranking when ``skipna`` is ``True``. The default is ``np.nan`` but can be set to any value to force NA values to the "bottom" or "top" of a rank as needed.''',
            axis='''axis: Integer specifying axis of ranking, where 0 ranks vertically (within each column) and 1 ranks horizontally (within each row)'''
            )
    reindex = dict(
            doc='''Return a new :obj:`{class_name}` with labels defined by the provided index. The size and ordering of the data is determined by the newly provided index, where data will continue to be aligned under labels found in both the new and the old index. Labels found only in the new index will be filled with ``fill_value``.
            ''',
            index_initializer=INDEX_INITIALIZER,
            fill_value='''fill_value: A value to be used to fill space created by a new index that has values not found in the previous index.''',
            own_index=OWN_INDEX,
            own_columns=OWN_COLUMNS
            )

    relabel = dict(
            doc ='''Return a new :obj:`{class_name}` with transformed labels on the index. The size and ordering of the data is never changed in a relabeling operation. The resulting index must be unique.
            ''',
            count='''A positive integer drops that many outer-most levels; a negative integer drops that many inner-most levels.''',
            level='''A hashable value to be used as a new root level, extending or creating an ``IndexHierarchy``''',
            relabel_input='''One of the following types, used to create a new ``Index`` with the same size as the previous index. (a) A mapping (as a dictionary or ``Series``), used to lookup and transform the labels in the previous index. Previous labels not found in the mapping will be reused. (b) A function, returning a hashable, that is applied to each label in the previous index. (c) The ``IndexAutoFactory`` type, to apply an auto-incremented integer index. (d) An index initializer, i.e., either an iterable of hashables or an ``Index`` instance.'''
            )

    relabel_flat = dict(
            doc='''Return a new :obj:`{class_name}`, where an ``IndexHierarchy`` (if defined) is replaced with a flat, one-dimension index of tuples.
            ''',
            )

    relabel_level_add = dict(
            doc='''Return a new :obj:`{class_name}`, adding a new root level to an existing ``IndexHierarchy``, or creating an ``IndexHierarchy`` if one is not yet defined.
            ''',
            level='''A hashable value to be used as a new root level, extending or creating an ``IndexHierarchy``''',
            )

    relabel_level_drop = dict(
            doc='''Return a new :obj:`{class_name}`, dropping one or more levels from a either the root or the leaves of an ``IndexHierarchy``. The resulting index must be unique.
            ''',
            count='''A positive integer drops that many outer-most (root) levels; a negative integer drops that many inner-most (leaf)levels.''',
            )
    sample = dict(
            doc='''Randomly (optionally made deterministic with a fixed seed) extract items from the container to return a subset of the container.''',
            count='''Number of elements to select.''',
            index='''Number of labels to select from the index.''',
            columns='''Number of labels to select from the columns.''',
            seed='''Initial state of random selection.''',
            )
    searchsorted = dict(
            doc='Given a sorted :obj:`Series`, return the {label_type} position(s) at which insertion in ``values`` would retain sort order.',
            values='values: a single value, or iterable of values.',
            side_left='side_left: If True, the index of the first suitable location found is given, else return the last such index. If matching an existing value, `side_left==True` will return that position, `side_left==Right` will return the next position (or the length).',
            fill_value='fill_value: A value to be used to fill the label beyond the last label.',
            )
    selector = dict(
            key_loc=LOC_SELECTOR,
            key_iloc=ILOC_SELECTOR,
            )
    sort = dict(
            ascending='If True, sort in ascending order; if False, sort in descending order.',
            kind='Name of the sort algorithm as passed to NumPy.',
            key='A function that is used to pre-process the selected columns or rows and derive new values to sort by.'
            )

    store_client_exporter = dict(
            args = f'''
        Args:
            {FP}
            {STORE_CONFIG_MAP}
            '''
            )

    tail = dict(
            doc='''Return a :obj:`{class_name}` consisting only of the bottom elements as specified by ``count``.
            ''',
            count='''count: Number of elements to be returned from the bottom of the :obj:`{class_name}`''',
            )


    ufunc_skipna = dict(
            args = '''
        Args:
            axis: Axis, defaulting to axis 0.
            skipna: Skip missing (NaN) values, defaulting to True.
            '''
            )
    window = dict(
            args = f'''
        Args:
            size: Elements per window, given as an integer greater than 0.
            {AXIS}
            step: Element shift per window, given as an integer greater than 0. Determines the step size between windows. A step of 1 shifts each window 1 element; a step equal to the ``size`` will result in non-overlapping windows.
            window_sized: if True, windows with fewer elements than ``size`` are skipped.
            window_func: Array processor of window values, executed before function application (if used): can be used for applying a weighting function to each window.
            window_valid: Function that, given an array window, returns True if the window is valid; invalid windows are skipped.
            label_shift: A shift, relative to the right-most element contained in the window, to derive the label to be paired with the window. For example, to label each window with the label found at the start of the window, ``label_shift`` can be set to one less than ``size``.
            start_shift: A shift to determine the first element where window collection begins.
            size_increment: A value to be added to ``size`` with each window after the first, so as to, in combination with setting ``step`` to 0, permit iterating over expanding windows.
            '''
            )

# NOTE: F here should replace AnyCallable below
F = tp.TypeVar('F', bound=tp.Callable[..., tp.Any])

def doc_inject(*,
        selector: tp.Optional[str] = None,
        **kwargs: object
        ) -> tp.Callable[[AnyCallable], AnyCallable]:
    '''
    Args:
        selector: optionally specify name of doc template dictionary to use; if not provided, the name of the function will be used.
    '''
    def decorator(f: AnyCallable) -> AnyCallable:

        assert f.__doc__ is not None, f'{f} must have a docstring!'

        nonlocal selector
        selector = f.__name__ if selector is None else selector
        # get doc string, template with decorator args, then template existing doc string
        doc_src = getattr(DOC_TEMPLATE, selector)
        if isinstance(doc_src, str):
            doc = doc_src.format(**kwargs)
            f.__doc__ = f.__doc__.format(doc)
        else: # assume it is a dictionary
            # try to format each value
            doc_dict = {k: v.format(**kwargs) for k, v in doc_src.items()}
            f.__doc__ = f.__doc__.format(**doc_dict)

        return f

    return decorator



'''
This module us for utilty functions that take as input and / or return Container subclasses such as Index, Series, or Frame, and that need to be shared by multiple such Container classes.
'''

from collections import defaultdict
from itertools import zip_longest
from functools import partial
import datetime
from fractions import Fraction
import typing as tp

import numpy as np
from numpy import char as npc
from arraykit import column_2d_filter

from static_frame.core.index_base import IndexBase
from static_frame.core.util import AnyCallable
from static_frame.core.util import Bloc2DKeyType
from static_frame.core.util import concat_resolved
from static_frame.core.util import DEFAULT_SORT_KIND
from static_frame.core.util import DepthLevelSpecifier
from static_frame.core.util import DTYPE_BOOL
from static_frame.core.util import DTYPE_OBJECT
from static_frame.core.util import DTYPE_STR
from static_frame.core.util import DTYPE_STR_KINDS
from static_frame.core.util import DtypesSpecifier
from static_frame.core.util import DtypeSpecifier
from static_frame.core.util import GetItemKeyType
from static_frame.core.util import IndexConstructor
from static_frame.core.util import IndexConstructors
from static_frame.core.util import IndexInitializer
from static_frame.core.util import iterable_to_array_1d
from static_frame.core.util import NULL_SLICE
from static_frame.core.util import slice_to_ascending_slice
from static_frame.core.util import STATIC_ATTR
from static_frame.core.util import UFunc
from static_frame.core.util import ufunc_set_iter
from static_frame.core.util import INT_TYPES
from static_frame.core.util import NameType
from static_frame.core.util import is_dtype_specifier
from static_frame.core.util import is_mapping

from static_frame.core.exception import AxisInvalid

if tp.TYPE_CHECKING:
    import pandas as pd #pylint: disable=W0611 #pragma: no cover
    from static_frame.core.type_blocks import TypeBlocks #pylint: disable=W0611,C0412 #pragma: no cover
    from static_frame.core.series import Series #pylint: disable=W0611,C0412 #pragma: no cover
    from static_frame.core.frame import Frame #pylint: disable=W0611,C0412 #pragma: no cover
    from static_frame.core.index_hierarchy import IndexHierarchy #pylint: disable=W0611,C0412 #pragma: no cover
    from static_frame.core.index_auto import IndexAutoFactoryType #pylint: disable=W0611,C0412 #pragma: no cover
    from static_frame.core.quilt import Quilt #pylint: disable=W0611,C0412 #pragma: no cover



def get_col_dtype_factory(
        dtypes: DtypesSpecifier,
        columns: tp.Optional[tp.Sequence[tp.Hashable]],
        ) -> tp.Callable[[int], np.dtype]:
    '''
    Return a function, or None, to get values from a DtypeSpecifier.

    Args:
        columns: In common usage in Frame constructors, ``columns`` is a reference to a mutable list that is assigned column labels when processing data (and before this function is called). Columns can also be an ``Index``.
    '''
    from static_frame.core.series import Series

    # dtypes are either a dtype initializer, mappable by name, or an ordered sequence
    # NOTE: might verify that all keys in dtypes are in columns, though that might be slow

    if is_mapping(dtypes):
        is_map = True
        is_element = False
    elif is_dtype_specifier(dtypes):
        is_map = False
        is_element = True
    else: # an iterable of types
        is_map = False
        is_element = False

    if columns is None and is_map:
        raise RuntimeError('cannot lookup dtypes by name without supplied columns labels')

    def get_col_dtype(col_idx: int) -> DtypeSpecifier:
        nonlocal dtypes # might mutate a generator into a tuple
        if is_element:
            return dtypes
        if is_map:
            return dtypes.get(columns[col_idx], None) #type: ignore
        # NOTE: dtypes might be a generator deferred until this function is called; if so, realize here; INVALID_ITERABLE_FOR_ARRAY (dict_values, etc) do not have __getitem__,
        if not hasattr(dtypes, '__len__') or not hasattr(dtypes, '__getitem__'):
            dtypes = tuple(dtypes) #type: ignore
        return dtypes[col_idx] #type: ignore

    return get_col_dtype


def is_static(value: IndexConstructor) -> bool:
    try:
        # if this is a class constructor
        return getattr(value, STATIC_ATTR) #type: ignore
    except AttributeError:
        pass
    # assume this is a class method
    return getattr(value.__self__, STATIC_ATTR) #type: ignore


def pandas_version_under_1() -> bool:
    import pandas
    return not hasattr(pandas, 'NA') # object introduced in 1.0

def pandas_to_numpy(
        container: tp.Union['pd.Index', 'pd.Series', 'pd.DataFrame'],
        own_data: bool,
        fill_value: tp.Any = np.nan
        ) -> np.ndarray:
    '''Convert Pandas container to a numpy array in pandas 1.0, where we might have Pandas extension dtypes that may have pd.NA. If no pd.NA, can go back to numpy types.

    If coming from a Pandas extension type, will convert pd.NA to `fill_value` in the resulting object array. For object dtypes, pd.NA may pass on into SF; the only way to find them is an expensive iteration and `is` comparison, which we are not sure we want to do at this time.

    Args:
        fill_value: if replcaing pd.NA, what to replace it with. Ultimately, this can use FillValueAuto to avoid going to object in all cases.
    '''
    # NOTE: only to be used with pandas 1.0 and greater

    if container.ndim == 1: # Series, Index
        dtype_src = container.dtype
        ndim = 1
    elif container.ndim == 2: # DataFrame, assume contiguous dtypes
        dtypes = container.dtypes.unique()
        assert len(dtypes) == 1
        dtype_src = dtypes[0]
        ndim = 2
    else:
        raise NotImplementedError(f'no handling for ndim {container.ndim}') #pragma: no cover

    if isinstance(dtype_src, np.dtype):
        dtype = dtype_src
        is_extension_dtype = False
    elif hasattr(dtype_src, 'numpy_dtype'):
        # only int, uint dtypes have this attribute
        dtype = dtype_src.numpy_dtype
        is_extension_dtype = True
    else:
        dtype = None # resolve below
        is_extension_dtype = True

    if is_extension_dtype:
        isna = container.isna() # returns a NumPy Boolean type sometimes
        if not isinstance(isna, np.ndarray):
            isna = isna.values
        hasna = isna.any() # will work for ndim 1 and 2

        from pandas import StringDtype #pylint: disable=E0611
        from pandas import BooleanDtype #pylint: disable=E0611
        # from pandas import DatetimeTZDtype
        # from pandas import Int8Dtype
        # from pandas import Int16Dtype
        # from pandas import Int32Dtype
        # from pandas import Int64Dtype
        # from pandas import UInt16Dtype
        # from pandas import UInt32Dtype
        # from pandas import UInt64Dtype
        # from pandas import UInt8Dtype

        if isinstance(dtype_src, BooleanDtype):
            dtype = DTYPE_OBJECT if hasna else DTYPE_BOOL
        elif isinstance(dtype_src, StringDtype):
            # trying to use a dtype argument for strings results in a converting pd.NA to a string "<NA>"
            dtype = DTYPE_OBJECT if hasna else DTYPE_STR
        else:
            # if an extension type and it hasna, have to go to object; otherwise, set to None or the dtype obtained above
            dtype = DTYPE_OBJECT if hasna else dtype

        # NOTE: in some cases passing the dtype might raise an exception, but it appears we are handling all of those cases by looking at hasna and selecting an object dtype
        array = container.to_numpy(copy=not own_data, dtype=dtype)

        if hasna:
            # if hasna and extension dtype, should be an object array; replace pd.NA objects with fill_value (np.nan)
            assert array.dtype == DTYPE_OBJECT
            array[isna] = fill_value

    else: # not an extension dtype
        if own_data:
            array = container.values
        else:
            array = container.values.copy()

    array.flags.writeable = False
    return array


def index_from_optional_constructor(
        value: IndexInitializer,
        *,
        default_constructor: IndexConstructor,
        explicit_constructor: tp.Optional[IndexConstructor] = None,
        ) -> IndexBase:
    '''
    Given a value that is an IndexInitializer (which means it might be an Index), determine if that value is really an Index, and if so, determine if a copy has to be made; otherwise, use the default_constructor. If an explicit_constructor is given, that is always used.
    '''
    # NOTE: this might return an own_index flag to show callers when a new index has been created

    if explicit_constructor:
        return explicit_constructor(value)

    # default constructor could be a function with a STATIC attribute
    if isinstance(value, IndexBase):
        # if default is STATIC, and value is not STATIC, get an immutabel
        if is_static(default_constructor):
            if not value.STATIC:
                # v: ~S, dc: S, use immutable alternative
                return value._IMMUTABLE_CONSTRUCTOR(value)
            # v: S, dc: S, both immutable
            return value
        else: # default constructor is mutable
            if not value.STATIC:
                # v: ~S, dc: ~S, both are mutable
                return value.copy()
            # v: S, dc: ~S, return a mutable version of something that is not mutable
            return value._MUTABLE_CONSTRUCTOR(value)

    # cannot always determine static status from constructors; fallback on using default constructor
    return default_constructor(value)

def index_constructor_empty(
        index: tp.Union[IndexInitializer, 'IndexAutoFactoryType']
        ) -> bool:
    '''
    Determine if an index is empty (if possible) or an IndexAutoFactory.
    '''
    from static_frame.core.index_auto import IndexAutoFactory
    if index is None or index is IndexAutoFactory:
        return True
    elif (not isinstance(index, IndexBase)
            and hasattr(index, '__len__')
            and len(index) == 0 #type: ignore
            ):
        return True
    return False

def matmul(
        lhs: tp.Union['Series', 'Frame', np.ndarray, tp.Sequence[float]],
        rhs: tp.Union['Series', 'Frame', np.ndarray, tp.Sequence[float]],
        ) -> tp.Any: #tp.Union['Series', 'Frame']:
    '''
    Implementation of matrix multiplication for Series and Frame
    '''
    from static_frame.core.series import Series
    from static_frame.core.frame import Frame

    # for a @ b = c
    # if a is 2D: a.columns must align b.index
    # if b is 1D, a.columns bust align with b.index
    # if a is 1D: len(a) == b.index (len of b), returns w columns of B

    if not isinstance(rhs, (np.ndarray, Series, Frame)):
        # try to make it into an array
        rhs = np.array(rhs)

    if not isinstance(lhs, (np.ndarray, Series, Frame)):
        # try to make it into an array
        lhs = np.array(lhs)

    if isinstance(lhs, np.ndarray):
        lhs_type = np.ndarray
    elif isinstance(lhs, Series):
        lhs_type = Series
    else: # normalize subclasses
        lhs_type = Frame

    if isinstance(rhs, np.ndarray):
        rhs_type = np.ndarray
    elif isinstance(rhs, Series):
        rhs_type = Series
    else: # normalize subclasses
        rhs_type = Frame

    if rhs_type == np.ndarray and lhs_type == np.ndarray:
        return np.matmul(lhs, rhs)


    own_index = True
    constructor = None

    if lhs.ndim == 1: # Series, 1D array
        # result will be 1D or 0D
        columns = None

        if lhs_type == Series and (rhs_type == Series or rhs_type == Frame):
            aligned = lhs._index.union(rhs._index)
            # if the aligned shape is not the same size as the originals, we do not have the same values in each and cannot proceed (all values go to NaN)
            if len(aligned) != len(lhs._index) or len(aligned) != len(rhs._index):
                raise RuntimeError('shapes not alignable for matrix multiplication') #pragma: no cover

        if lhs_type == Series:
            if rhs_type == np.ndarray:
                if lhs.shape[0] != rhs.shape[0]: # works for 1D and 2D
                    raise RuntimeError('shapes not alignable for matrix multiplication')
                ndim = rhs.ndim - 1 # if 2D, result is 1D, of 1D, result is 0
                left = lhs.values
                right = rhs # already np
                if ndim == 1:
                    index = None # force auto increment integer
                    own_index = False
                    constructor = lhs.__class__
            elif rhs_type == Series:
                ndim = 0
                left = lhs.reindex(aligned).values
                right = rhs.reindex(aligned).values
            else: # rhs is Frame
                ndim = 1
                left = lhs.reindex(aligned).values
                right = rhs.reindex(index=aligned).values
                index = rhs._columns
                constructor = lhs.__class__
        else: # lhs is 1D array
            left = lhs
            right = rhs.values
            if rhs_type == Series:
                ndim = 0
            else: # rhs is Frame, len(lhs) == len(rhs.index)
                ndim = 1
                index = rhs._columns
                constructor = Series # cannot get from argument

    elif lhs.ndim == 2: # Frame, 2D array

        if lhs_type == Frame and (rhs_type == Series or rhs_type == Frame):
            aligned = lhs._columns.union(rhs._index)
            # if the aligned shape is not the same size as the originals, we do not have the same values in each and cannot proceed (all values go to NaN)
            if len(aligned) != len(lhs._columns) or len(aligned) != len(rhs._index):
                raise RuntimeError('shapes not alignable for matrix multiplication')

        if lhs_type == Frame:
            if rhs_type == np.ndarray:
                if lhs.shape[1] != rhs.shape[0]: # works for 1D and 2D
                    raise RuntimeError('shapes not alignable for matrix multiplication')
                ndim = rhs.ndim
                left = lhs.values
                right = rhs # already np
                index = lhs._index

                if ndim == 1:
                    constructor = Series
                else:
                    constructor = lhs.__class__
                    columns = None # force auto increment index
            elif rhs_type == Series:
                # a.columns must align with b.index
                ndim = 1
                left = lhs.reindex(columns=aligned).values
                right = rhs.reindex(aligned).values
                index = lhs._index  # this axis is not changed
                constructor = rhs.__class__
            else: # rhs is Frame
                # a.columns must align with b.index
                ndim = 2
                left = lhs.reindex(columns=aligned).values
                right = rhs.reindex(index=aligned).values
                index = lhs._index
                columns = rhs._columns
                constructor = lhs.__class__ # give left precedence
        else: # lhs is 2D array
            left = lhs
            right = rhs.values
            if rhs_type == Series: # returns unindexed Series
                ndim = 1
                index = None
                own_index = False
                constructor = rhs.__class__
            else: # rhs is Frame, lhs.shape[1] == rhs.shape[0]
                if lhs.shape[1] != rhs.shape[0]: # works for 1D and 2D
                    raise RuntimeError('shapes not alignable for matrix multiplication')
                ndim = 2
                index = None
                own_index = False
                columns = rhs._columns
                constructor = rhs.__class__
    else:
        raise NotImplementedError(f'no handling for {lhs}')

    # NOTE: np.matmul is not the same as np.dot for some arguments
    data = np.matmul(left, right)

    if ndim == 0:
        return data

    assert constructor is not None

    data.flags.writeable = False
    if ndim == 1:
        return constructor(data,
                index=index,
                own_index=own_index,
                )
    return constructor(data,
            index=index,
            own_index=own_index,
            columns=columns
            )


def axis_window_items( *,
        source: tp.Union['Series', 'Frame', 'Quilt'],
        size: int,
        axis: int = 0,
        step: int = 1,
        window_sized: bool = True,
        window_func: tp.Optional[AnyCallable] = None,
        window_valid: tp.Optional[AnyCallable] = None,
        label_shift: int = 0,
        start_shift: int = 0,
        size_increment: int = 0,
        as_array: bool = False,
        ) -> tp.Iterator[tp.Tuple[tp.Hashable, tp.Any]]:
    '''Generator of index, window pairs. When ndim is 2, axis 0 returns windows of rows, axis 1 returns windows of columns.

    Args:
        as_array: if True, the window is returned as an array instead of a SF object.
    '''
    # see doc_str window for docs

    from static_frame.core.frame import Frame
    from static_frame.core.quilt import Quilt
    from static_frame.core.series import Series

    if size <= 0:
        raise RuntimeError('window size must be greater than 0')
    if step < 0:
        raise RuntimeError('window step cannot be less than than 0')

    source_ndim = source.ndim
    values: tp.Optional[np.ndarray] = None

    if source_ndim == 1:
        assert isinstance(source, Series) # for mypy
        labels = source._index
        if as_array:
            values = source.values
    else:
        labels = source._index if axis == 0 else source._columns #type: ignore

        if isinstance(source, Frame) and axis == 0 and as_array:
            # for a Frame, when collecting rows, it is more efficient to pre-consolidate blocks prior to slicing. Note that this results in the same block coercion necessary for each window (which is not the same for axis 1, where block coercion is not required)
            values = source._blocks.values

    if start_shift >= 0:
        count_window_max = len(labels)
    else: # add for iterations when less than 0
        count_window_max = len(labels) + abs(start_shift)

    idx_left_max = count_window_max - 1
    idx_left = start_shift
    count = 0

    while True:
        # idx_left, size can change over iterations
        idx_right = idx_left + size - 1

        # floor idx_left at 0 so as to not wrap
        idx_left_floored = idx_left if idx_left > 0 else 0
        idx_right_floored = idx_right if idx_right > -1 else -1 # will add one

        key = slice(idx_left_floored, idx_right_floored + 1)

        if source_ndim == 1:
            if as_array:
                window = values[key] #type: ignore
            else:
                window = source._extract_iloc(key)
        else:
            if axis == 0: # extract rows
                if as_array and values is not None:
                    window = values[key]
                elif as_array:
                    window = source._extract_array(key) #type: ignore
                else: # use low level iloc selector
                    window = source._extract(row_key=key) #type: ignore
            else: # extract columns
                if as_array:
                    window = source._extract_array(NULL_SLICE, key) #type: ignore
                else:
                    window = source._extract(column_key=key) #type: ignore

        valid = True
        try:
            idx_label = idx_right + label_shift
            if idx_label < 0: # do not wrap around
                raise IndexError()
            #if we cannot get a label, the window is invalid
            label = labels.iloc[idx_label]
        except IndexError: # an invalid label has to be dropped
            valid = False

        if valid and window_sized and window.shape[axis] != size:
            valid = False
        if valid and window_valid and not window_valid(window):
            valid = False

        if valid:
            if window_func:
                window = window_func(window)
            yield label, window

        idx_left += step
        size += size_increment
        count += 1

        if count > count_window_max or idx_left > idx_left_max or size < 0:
            break

def get_block_match(
        width: int,
        values_source: tp.List[np.ndarray],
        ) -> tp.Iterator[np.ndarray]:
    '''Utility method for assignment. Draw from values to provide as many columns as specified by width. Use `values_source` as a stack to draw and replace values.
    '''
    # see clip().get_block_match() for one example of drawing values from another sequence of blocks, where we take blocks and slices from blocks using a list as a stack

    if width == 1: # no loop necessary
        v = values_source.pop()
        if v.ndim == 1:
            yield v
        else: # ndim == 2
            if v.shape[1] > 1: # more than one column
                # restore remained to values source
                values_source.append(v[NULL_SLICE, 1:])
            yield v[NULL_SLICE, 0]
    else:
        width_found = 0
        while width_found < width:
            v = values_source.pop()
            if v.ndim == 1:
                yield v
                width_found += 1
                continue
            # ndim == 2
            width_v = v.shape[1]
            width_needed = width - width_found
            if width_v <= width_needed:
                yield v
                width_found += width_v
                continue
            # width_v > width_needed
            values_source.append(v[NULL_SLICE, width_needed:])
            yield v[NULL_SLICE, :width_needed]
            break

def bloc_key_normalize(
        key: Bloc2DKeyType,
        container: 'Frame'
        ) -> np.ndarray:
    '''
    Normalize and validate a bloc key. Return a same sized Boolean array.
    '''
    from static_frame.core.frame import Frame

    if isinstance(key, Frame):
        bloc_frame = key.reindex(
                index=container._index,
                columns=container._columns,
                fill_value=False
                )
        bloc_key = bloc_frame.values # shape must match post reindex
    elif key.__class__ is np.ndarray:
        bloc_key = key
        if bloc_key.shape != container.shape:
            raise RuntimeError(f'bloc {bloc_key.shape} must match shape {container.shape}')
    else:
        raise RuntimeError(f'invalid bloc_key, must be Frame or array, not {key}')

    if not bloc_key.dtype == bool:
        raise RuntimeError('cannot use non-Boolean dtype as bloc key')

    return bloc_key


def key_to_ascending_key(key: GetItemKeyType, size: int) -> GetItemKeyType:
    '''
    Normalize all types of keys into an ascending formation.

    Args:
        size: the length of the container on this axis
    '''
    from static_frame.core.frame import Frame
    from static_frame.core.series import Series

    if key.__class__ is slice:
        return slice_to_ascending_slice(key, size=size) #type: ignore

    if isinstance(key, str) or not hasattr(key, '__len__'):
        return key

    if key.__class__ is np.ndarray:
        # array first as not truthy
        if key.dtype == bool: #type: ignore
            return key
        return np.sort(key, kind=DEFAULT_SORT_KIND)

    if not len(key): #type: ignore
        return key

    if isinstance(key, list):
        return sorted(key)

    if isinstance(key, Series):
        return key.sort_index()

    if isinstance(key, Frame):
        # for usage in assignment we need columns to be sorted
        return key.sort_columns()

    raise RuntimeError(f'unhandled key {key}')


def rehierarch_from_type_blocks(*,
        labels: 'TypeBlocks',
        depth_map: tp.Sequence[int],
        index_cls: tp.Type['IndexHierarchy'],
        index_constructors: tp.Optional[IndexConstructors] = None,
        name: tp.Optional[tp.Hashable] = None,
        ) -> tp.Tuple['IndexBase', np.ndarray]:
    '''
    Given labels suitable for a hierarchical index, order them into a hierarchy using the given depth_map.

    Args:
        index_cls: provide a class, from which the constructor will be called.
    '''

    depth = labels.shape[1] # number of columns

    if depth != len(depth_map):
        raise RuntimeError('must specify new depths for all depths')
    if set(range(depth)) != set(depth_map):
        raise RuntimeError('all depths must be specified')

    labels_post = labels._extract(row_key=NULL_SLICE, column_key=list(depth_map))
    labels_sort = np.full(labels_post.shape, 0)

    # get ordering of values found in each level
    order: tp.List[tp.Dict[tp.Hashable, int]] = [defaultdict(int) for _ in range(depth)]

    for (idx_row, idx_col), label in labels.element_items():
        if label not in order[idx_col]:
            # Map label to an integer representing the observed order.
            order[idx_col][label] = len(order[idx_col])
        # Fill array for sorting based on observed order.
        labels_sort[idx_row, idx_col] = order[idx_col][label]

    # Reverse depth_map for lexical sorting, which sorts by rightmost column first.
    order_lex = np.lexsort(
            [labels_sort[NULL_SLICE, i] for i in reversed(depth_map)])

    labels_post = labels_post._extract(row_key=order_lex)

    index = index_cls._from_type_blocks(
            blocks=labels_post,
            index_constructors=index_constructors,
            name=name,
            own_blocks=True,
            )
    return index, order_lex

def rehierarch_from_index_hierarchy(*,
        labels: 'IndexHierarchy',
        depth_map: tp.Sequence[int],
        index_constructors: tp.Optional[IndexConstructors] = None,
        name: tp.Optional[tp.Hashable] = None,
        ) -> tp.Tuple['IndexBase', np.ndarray]:
    '''
    Alternate interface that updates IndexHierarchy cache before rehierarch.
    '''
    if labels._recache:
        labels._update_array_cache()

    return rehierarch_from_type_blocks(
            labels=labels._blocks,
            depth_map=depth_map,
            index_cls=labels.__class__,
            index_constructors=index_constructors,
            name=name,
            )

def array_from_value_iter(
        key: tp.Hashable,
        idx: int,
        get_value_iter: tp.Callable[[tp.Hashable], tp.Iterator[tp.Any]],
        get_col_dtype: tp.Optional[tp.Callable[[int], np.dtype]],
        row_count: int,
        ) -> np.ndarray:
    '''
    Return a single array given keys and collections.

    Args:
        get_value_iter: Iterator of a values
        dtypes: if an
        key: hashable for looking up field in `get_value_iter`.
        idx: integer position to extract from dtypes
    '''
    # for each column, try to get a dtype, or None
    # if this value is None we cannot tell if it was explicitly None or just was not specified
    dtype = None if get_col_dtype is None else get_col_dtype(idx)

    # NOTE: shown to be faster to try fromiter in some performance tests
    # values, _ = iterable_to_array_1d(get_value_iter(key), dtype=dtype)

    values = None
    if dtype is not None:
        try:
            values = np.fromiter(
                    get_value_iter(key),
                    count=row_count,
                    dtype=dtype)
            values.flags.writeable = False
        except (ValueError, TypeError):
            # the dtype may not be compatible, so must fall back on using np.array to determine the type, i.e., ValueError: cannot convert float NaN to integer
            pass
    if values is None:
        # returns an immutable array
        values, _ = iterable_to_array_1d(
                get_value_iter(key),
                dtype=dtype
                )
    return values

#-------------------------------------------------------------------------------
# utilities for binary operator applications with type blocks

def apply_binary_operator(*,
        values: np.ndarray,
        other: tp.Any,
        other_is_array: bool,
        operator: UFunc,
        ) -> np.ndarray:
    '''
    Utility to handle binary operator application.
    '''
    if (values.dtype.kind in DTYPE_STR_KINDS or
            (other_is_array and other.dtype.kind in DTYPE_STR_KINDS)):
        operator_name = operator.__name__

        if operator_name == 'add':
            result = npc.add(values, other)
        elif operator_name == 'radd':
            result = npc.add(other, values)
        elif operator_name == 'mul' or operator_name == 'rmul':
            result = npc.multiply(values, other)
        else:
            result = operator(values, other)
    else:
        result = operator(values, other)

    if result is False or result is True:
        if not other_is_array and (
                isinstance(other, str) or not hasattr(other, '__len__')
                ):
            # only expand to the size of the array operand if we are comparing to an element
            result = np.full(values.shape, result, dtype=DTYPE_BOOL)
        elif other_is_array and other.size == 1:
            # elements in arrays of 0 or more dimensions are acceptable; this is what NP does for arithmetic operators when the types are compatible
            result = np.full(values.shape, result, dtype=DTYPE_BOOL)
        else:
            raise ValueError('operands could not be broadcast together')
            # raise on unaligned shapes as is done for arithmetic operators

    result.flags.writeable = False
    return result

def apply_binary_operator_blocks(*,
        values: tp.Iterable[np.ndarray],
        other: tp.Iterable[np.ndarray],
        operator: UFunc,
        apply_column_2d_filter: bool,
    ) -> tp.Iterator[np.ndarray]:
    '''
    Application from iterators of arrays, to iterators of arrays.
    '''
    if apply_column_2d_filter:
        values = (column_2d_filter(op) for op in values)
        other = (column_2d_filter(op) for op in other)

    for a, b in zip_longest(values, other):
        yield apply_binary_operator(
                values=a,
                other=b,
                other_is_array=True,
                operator=operator,
                )

def apply_binary_operator_blocks_columnar(*,
        values: tp.Iterable[np.ndarray],
        other: np.ndarray,
        operator: UFunc,
    ) -> tp.Iterator[np.ndarray]:
    '''
    Application from iterators of arrays, to iterators of arrays. Will return iterator of all 1D arrays, as we will break down larger blocks in values into 1D arrays.

    Args:
        other: 1D array to be applied to each column of the blocks.
    '''
    assert other.ndim == 1
    for block in values:
        if block.ndim == 1:
            yield apply_binary_operator(
                    values=block,
                    other=other,
                    other_is_array=True,
                    operator=operator,
                    )
        else:
            for i in range(block.shape[1]):
                yield apply_binary_operator(
                        values=block[NULL_SLICE, i],
                        other=other,
                        other_is_array=True,
                        operator=operator,
                        )

#-------------------------------------------------------------------------------

def arrays_from_index_frame(
        container: 'Frame',
        depth_level: tp.Optional[DepthLevelSpecifier],
        columns: GetItemKeyType
        ) -> tp.Iterator[np.ndarray]:
    '''
    Given a Frame, return an iterator of index and / or columns as 1D or 2D arrays.
    '''
    if depth_level is not None:
        # NOTE: a 1D index of tuples will be taken as a 1D array of tuples; there is no obvious way to treat this as 2D array without guessing that we are trying to match an IndexHierarchy
        # NOTE: if a multi-column selection, might be better to yield one depth at a time
        yield container.index.values_at_depth(depth_level)
    if columns is not None:
        column_key = container.columns._loc_to_iloc(columns)
        yield from container._blocks._slice_blocks(column_key=column_key)


def key_from_container_key(
        index: IndexBase,
        key: GetItemKeyType,
        expand_iloc: bool = False,
        ) -> GetItemKeyType:
    '''
    Unpack selection values from another Index, Series, or ILoc selection.
    '''
    # PERF: do not do comparisons if key is not a Container or SF object
    if not hasattr(key, 'STATIC'):
        return key

    from static_frame.core.index import Index
    from static_frame.core.index import ILoc
    from static_frame.core.series import Series
    from static_frame.core.series import SeriesHE

    if isinstance(key, Index):
        # if an Index, we simply use the values of the index
        key = key.values
    elif isinstance(key, Series) and key.__class__ is not SeriesHE:
        # Series that are not hashable are unpacked into an array; SeriesHE can be used as a key
        if key.dtype == DTYPE_BOOL:
            # if a Boolean series, sort and reindex
            if not key.index.equals(index):
                key = key.reindex(index,
                        fill_value=False,
                        check_equals=False,
                        ).values
            else: # the index is equal
                key = key.values
        else:
            # For all other Series types, we simply assume that the values are to be used as keys in the IH. This ignores the index, but it does not seem useful to require the Series, used like this, to have a matching index value, as the index and values would need to be identical to have the desired selection.
            key = key.values
    elif expand_iloc and key.__class__ is ILoc:
        # realize as Boolean array
        array = np.full(len(index), False)
        array[key.key] = True #type: ignore
        key = array

    # detect and fail on Frame?
    return key


#---------------------------------------------------------------------------

def _index_many_to_one(
        indices: tp.Iterable[IndexBase],
        cls_default: tp.Type[IndexBase],
        array_processor: tp.Callable[[tp.Iterable[np.ndarray]], np.ndarray]
        ) -> IndexBase:
    '''
    Given multiple Index objects, combine them. Preserve name and index type if aligned, and handle going to GO if the default class is GO.

    Args:
        indices: can be a generator
        cls_default: Default Index class to be used if no alignment of classes; also used to determine if result Index should be static or mutable.
    '''
    indices_iter = iter(indices)
    try:
        index = next(indices_iter)
    except StopIteration:
        return cls_default.from_labels(())

    arrays = [index.values]

    name_first = index.name
    name_aligned = True

    cls_first = index.__class__
    cls_aligned = True

    # if IndexHierarchy, collect index_types generators
    if index.ndim == 2:
        depth_first = index.depth
        index_types_gen = [index._levels.index_types()] #type: ignore
        index_types_aligned = True
    else:
        index_types_aligned = False

    for index in indices_iter:
        arrays.append(index.values)
        if name_aligned and index.name != name_first:
            name_aligned = False
        if cls_aligned and index.__class__ != cls_first:
            cls_aligned = False

        if index_types_aligned and index.ndim == 2 and index.depth == depth_first:
            index_types_gen.append(index._levels.index_types()) #type: ignore
        else:
            index_types_aligned = False

    if index_types_aligned:
        # all depths are already aligned
        index_constructors = []
        for types in zip(*index_types_gen):
            if all(types[0] == t for t in types[1:]):
                index_constructors.append(types[0])
            else: # assume this is always a 1D index
                index_constructors.append(cls_default)

    if cls_aligned:
        if cls_default.STATIC and not cls_first.STATIC:
            # default is static but aligned is mutable
            constructor = cls_first._IMMUTABLE_CONSTRUCTOR.from_labels #type: ignore
        elif not cls_default.STATIC and cls_first.STATIC:
            # default is mutable but aligned is static
            constructor = cls_first._MUTABLE_CONSTRUCTOR.from_labels #type: ignore
        else:
            constructor = cls_first.from_labels
    else:
        constructor = cls_default.from_labels

    name = name_first if name_aligned else None

    # returns an immutable array
    array = array_processor(arrays)

    if index_types_aligned:
        return constructor(array, name=name, index_constructors=index_constructors) #type: ignore
    return constructor(array, name=name) #type: ignore

def index_many_concat(
        indices: tp.Iterable[IndexBase],
        cls_default: tp.Type[IndexBase],
        ) -> tp.Optional[IndexBase]:
    return _index_many_to_one(indices, cls_default, concat_resolved)

def index_many_set(
        indices: tp.Iterable[IndexBase],
        cls_default: tp.Type[IndexBase],
        union: bool,
        ) -> tp.Optional[IndexBase]:
    '''
    Given multiple Index objects, concatenate them in order. Preserve name and index type if aligned.
    '''
    array_processor = partial(ufunc_set_iter,
            union=union,
            assume_unique=True)
    return _index_many_to_one(indices, cls_default, array_processor)


#-------------------------------------------------------------------------------
def apex_to_name(
        rows: tp.Sequence[tp.Sequence[tp.Hashable]],
        depth_level: tp.Optional[DepthLevelSpecifier],
        axis: int, # 0 is by row (for index), 1 is by column (for columns)
        axis_depth: int,
        ) -> NameType:
    '''
    Utility for translating apex values (the upper left corner created be index/columns) in the appropriate name.
    '''
    if depth_level is None:
        return None
    if axis == 0:
        if isinstance(depth_level, INT_TYPES):
            row = rows[depth_level]
            if axis_depth == 1: # return a single label
                return row[0] if row[0] != '' else None
            else:
                return tuple(row)
        else: # its a list selection
            targets = [rows[level] for level in depth_level]
            # combine into tuples
            if axis_depth == 1:
                return next(zip(*targets))
            else:
                return tuple(zip(*targets))
    elif axis == 1:
        if isinstance(depth_level, INT_TYPES):
            # depth_level refers to position in inner row
            row = [r[depth_level] for r in rows]
            if axis_depth == 1: # return a single label
                return row[0] if row[0] != '' else None
            else:
                return tuple(row)
        else: # its a list selection
            targets = (tuple(row[level] for level in depth_level) for row in rows) #type: ignore
            # combine into tuples
            if axis_depth == 1:
                return next(targets) #type: ignore
            else:
                return tuple(targets)

    raise AxisInvalid(f'invalid axis: {axis}')


def container_to_exporter_attr(container_type: tp.Type['Frame']) -> str:
    from static_frame.core.frame import Frame
    from static_frame.core.frame import FrameGO
    from static_frame.core.frame import FrameHE

    if container_type is Frame:
        return 'to_frame'
    elif container_type is FrameGO:
        return 'to_frame_go'
    elif container_type is FrameHE:
        return 'to_frame_he'
    raise NotImplementedError(f'no handling for {container_type}')


def sort_index_for_order(
        index: IndexBase,
        ascending: bool,
        kind: str,
        key: tp.Optional[tp.Callable[[IndexBase], tp.Union[np.ndarray, IndexBase]]],
        ) -> np.ndarray:
    '''Return an integer array defing the new ordering.
    '''
    # cfs is container_for_sort
    if key:
        cfs = key(index)
        cfs_is_array = cfs.__class__ is np.ndarray
        if cfs_is_array:
            cfs_depth = 1 if cfs.ndim == 1 else cfs.shape[1]
        else:
            cfs_depth = cfs.depth
        if len(cfs) != len(index):
            raise RuntimeError('key function returned a container of invalid length')
    else:
        cfs = index
        cfs_is_array = False
        cfs_depth = cfs.depth

    # argsort lets us do the sort once and reuse the results
    if cfs_depth > 1:
        if cfs_is_array:
            values_for_lex = [cfs[NULL_SLICE, i] for i in range(cfs.shape[1]-1, -1, -1)]
        else: # cfs is an IndexHierarchy
            values_for_lex = [cfs.values_at_depth(i)
                    for i in range(cfs.depth-1, -1, -1)]
        order = np.lexsort(values_for_lex)
    else:
        # depth is 1
        v = cfs if cfs_is_array else cfs.values
        order = np.argsort(v, kind=kind)

    if not ascending:
        order = order[::-1]
    return order

#-------------------------------------------------------------------------------

class MessagePackElement:
    '''
    Handle encoding/decoding of elements found in object arrays not well supported by msgpack. Many of these cases were found through Hypothesis testing.
    '''

    @staticmethod
    def encode(
            a: tp.Any,
            packb: AnyCallable,
            ) -> tp.Tuple[str, tp.Any]:

        if isinstance(a, datetime.datetime): #msgpack-numpy has an issue with datetime
            year = str(a.year).zfill(4) #datetime returns inconsistent year string for <4 digit years on some systems
            d = year + ' ' + a.strftime('%a %b %d %H:%M:%S:%f')
            return ('DT', d)
        elif isinstance(a, datetime.date):
            year = str(a.year).zfill(4) #datetime returns inconsistent year string for <4 digit years on some systems
            d = year + ' ' + a.strftime('%a %b %d')
            return ('D', d)
        elif isinstance(a, datetime.time):
            return ('T', a.strftime('%H:%M:%S:%f'))
        elif isinstance(a, np.ndarray): #recursion not covered by msgpack-numpy
            return ('A', packb(a)) #recurse packb
        elif isinstance(a, Fraction): #msgpack-numpy has an issue with fractions
            return ('F',  str(a))
        elif isinstance(a, int) and len(str(a)) >=19:
            #msgpack-python has an overflow issue with large ints
            return ('I', str(a))
        return ('', a)


    @staticmethod
    def decode(
            pair: tp.Tuple[str, tp.Any],
            unpackb: AnyCallable,
            ) -> tp.Any:
        dt = datetime.datetime

        (typ, d) = pair
        if typ == 'DT': #msgpack-numpy has an issue with datetime
            return dt.strptime(d, '%Y %a %b %d %H:%M:%S:%f')
        elif typ == 'D':
            return dt.strptime(d, '%Y %a %b %d').date()
        elif typ == 'T':
            return dt.strptime(d, '%H:%M:%S:%f').time()
        elif typ == 'F': #msgpack-numpy has an issue with fractions
            return Fraction(d)
        elif typ == 'I': #msgpack-python has an issue with very large int values
            return int(d)
        elif typ == 'A': #recursion not covered by msgpack-numpy
            return unpackb(d) #recurse unpackb
        return d


























import unittest
import frame_fixtures as ff


from static_frame import Series
from static_frame import SeriesHE
from static_frame.test.test_case import TestCase
from static_frame import ILoc

class TestUnit(TestCase):

    def test_frame_he_slotted_a(self) -> None:

        f1 = SeriesHE.from_element(1, index=(1,2))

        with self.assertRaises(AttributeError):
            f1.g = 30 #type: ignore #pylint: disable=E0237
        with self.assertRaises(AttributeError):
            f1.__dict__ #pylint: disable=W0104


    #---------------------------------------------------------------------------
    def test_series_he_hash_a(self) -> None:

        s1 = SeriesHE(('a', 'b', 'c'))
        s2 = SeriesHE(('a', 'b', 'c'))
        s3 = SeriesHE(('a', 'b', 'c'), index=tuple('xyz'))
        s4 = SeriesHE(('p', 'q', 'r'))
        s5 = Series(('a', 'b', 'c'))

        self.assertTrue(s1 == s2)
        # we do not compare container class, so
        self.assertTrue(s1 == s5)
        self.assertTrue(isinstance(s5 == s5, Series))

        self.assertTrue(isinstance(hash(s1), int))
        self.assertEqual(hash(s1), hash(s2))

        # has is based on index labels
        self.assertEqual(hash(s1), hash(s4))
        self.assertNotEqual(hash(s1), hash(s3))

        q = {s1, s2, s3}
        self.assertEqual(len(q), 2)
        self.assertTrue(s1 in q)
        self.assertTrue(s2 in q)
        self.assertTrue(s4 not in q)
        with self.assertRaises(TypeError):
            _ = s5 in q

        d = {s1: 'foo', s3: 'bar'}
        self.assertTrue(s2 in d)
        self.assertEquals(d[s1], 'foo')

        with self.assertRaises(TypeError):
            _ = s5 in d

        s6 = s1.to_series()

        with self.assertRaises(TypeError):
            _ = s6 in d


    def test_series_he_hash_b(self) -> None:

        s1 = SeriesHE.from_dict(dict(a=10, b=42))
        s2 = SeriesHE.from_dict(dict(x='foo', y='bar'), name=s1)

        self.assertEqual(s2.name['b'], 42)

        s3 = Series.from_dict({s1: 40, s2: 1000})

        self.assertEqual(s3[s2], 1000)
        self.assertFalse(s1 == s2) # name attr is different


    def test_series_he_hash_c(self) -> None:

        s1 = ff.parse('s(10,1)|i(I,str)')[ILoc[0]].to_series_he()
        self.assertFalse(hasattr(s1, '_hash'))
        self.assertEqual(hash(s1), s1._hash)


if __name__ == '__main__':
    unittest.main()





import typing as tp
import itertools as it
import string
import hashlib

import pandas as pd
import numpy as np

import static_frame as sf

from static_frame.performance.perf_test import PerfTest

#-------------------------------------------------------------------------------

def get_sample_series_float(size: int = 10000) -> tp.Tuple[pd.Series, sf.Series, np.ndarray]:
    a1 = np.arange(size) * .001
    a1[size // 2:] = np.nan
    pds = pd.Series(a1)
    sfs = sf.Series(a1)
    return pds, sfs, a1

def get_sample_series_string_index_float_values(size: int = 10000) -> tp.Tuple[pd.Series, sf.Series, np.ndarray]:
    a1 = np.arange(size) * .001
    a1[size // 2:] = np.nan
    # create hsa indices
    index = [hashlib.sha224(str(x).encode('utf-8')).hexdigest() for x in range(size)]
    pds = pd.Series(a1, index=index)
    sfs = sf.Series(a1, index=index)
    return pds, sfs, a1

def get_sample_series_obj(size: int = 10000) -> tp.Tuple[pd.Series, sf.Series, np.ndarray]:
    sample = [None, 3, 0.123, np.nan]
    a1 = np.array(sample * int(size / len(sample)))

    pds = pd.Series(a1)
    sfs = sf.Series(a1)

    return pds, sfs, a1

def get_sample_series_objstr(size: int = 10000) -> tp.Tuple[pd.Series, sf.Series, np.ndarray]:
    sample = [None, 3, 0.123, np.nan, 'str']
    a1 = np.array(sample * int(size / len(sample)))

    pds = pd.Series(a1)
    sfs = sf.Series(a1)

    return pds, sfs, a1


#-------------------------------------------------------------------------------
# frame generators


def get_sample_frame_float_string_index(
        size: int = 10000,
        columns: int = 100
        ) -> tp.Tuple[pd.DataFrame, sf.Frame, np.ndarray]:
    a1 = (np.arange(size * columns)).reshape((size, columns)) * .001
    # insert random nan in very other columns
    for col in range(0, 100, 2):
        a1[:100, col] = np.nan
    index = [hashlib.sha224(str(x).encode('utf-8')).hexdigest() for x in range(size)]
    cols = [hashlib.sha224(str(x).encode('utf-8')).hexdigest() for x in range(columns)]
    sff = sf.Frame(a1, index=index, columns=cols)
    pdf = pd.DataFrame(a1, index=index, columns=cols)
    return pdf, sff, a1


_mixed_types = ('foo', 'bar', True, None, 234.34, 90)

def _typed_array(dtype: type, size: int, shift: int = 0) -> np.ndarray:
    if dtype == float:
        return np.roll(np.arange(size) * .001, shift)
    if dtype == int:
        return np.roll(np.arange(size), shift)
    if dtype == bool:
        return np.roll(np.isin(np.arange(size) % 5, (1, 4)), shift)
    if dtype == object:
        return np.roll(np.array([_mixed_types[x % len(_mixed_types)] for x in range(size)]), shift)
    raise NotImplementedError()

def get_sample_frame_mixed_string_index(size: int = 10000, columns: int = 100) -> tp.Tuple[
            pd.DataFrame, sf.FrameGO, np.ndarray]:
    '''Get frames with mixed types.
    '''
    # produces 14950 strings
    source_ids = list(''.join(x) for x in it.combinations(string.ascii_lowercase, 4))
    assert size <= len(source_ids)

    index = source_ids[:size]
    cols = source_ids[:columns]

    dtypes = (float, int, object, bool)

    sff = sf.FrameGO(index=index)
    for idx, col in enumerate(cols):
        s = sf.Series(_typed_array(dtypes[idx % 4], size=size, shift=idx), index=index)
        sff[col] = s

    npf = sff.values

    pdf = pd.DataFrame(index=index)
    for idx, col in enumerate(cols):
        s = pd.Series(_typed_array(dtypes[idx % 4], size=size, shift=idx), index=index)
        pdf[col] = s


    return pdf, sff, npf


def get_series_float_h2d_str_index(size: int = 1000) -> tp.Tuple[pd.Series, sf.Series]:
    '''
    Get a hierarchical index with
    '''
    labels = list(''.join(x) for x in it.combinations(string.ascii_lowercase, 4))
    labels0 = labels[:int(size / 10)]
    labels1 = labels[:size]
    values = np.arange(len(labels0) * len(labels1)) * .001

    ih = sf.IndexHierarchy.from_product(labels0, labels1)
    sfs = sf.Series(values, index=ih)

    mi = pd.MultiIndex.from_product((labels0, labels1))
    pds = pd.Series(values, index=mi)
    return pds, sfs


def get_series_float_h3d_str_index(size: int = 1000) -> tp.Tuple[pd.Series, sf.Series]:
    '''
    Get a hierarchical index with
    '''
    labels = list(''.join(x) for x in it.combinations(string.ascii_lowercase, 4))
    labels0 = labels[:int(size / 100)]
    labels1 = labels[:int(size / 10)]
    labels2 = labels[:size]

    values = np.arange(len(labels0) * len(labels1) * len(labels2)) * .001

    ih = sf.IndexHierarchy.from_product(labels0, labels1, labels2)
    sfs = sf.Series(values, index=ih)

    mi = pd.MultiIndex.from_product((labels0, labels1, labels2))
    pds = pd.Series(values, index=mi)
    return pds, sfs

class SampleData:

    _store: tp.Dict[str, tp.Any] = {}

    @classmethod
    def create(cls) -> None:
        pds_int_float_10k, sfs_int_float_10k, npa_int_float_10k = get_sample_series_float(10000)
        pds_obj_10k, sfs_obj_10k, npa_obj_10k = get_sample_series_obj(10000)
        pds_str_float_10k, sfs_str_float_10k, npa_str_float_10k = get_sample_series_string_index_float_values(10000)
        pds_objstr_10k, sfs_objstr_10k, npa_objstr_10k = get_sample_series_objstr(10000)
        pdf_float_10k, sff_float_10k, npf_float_10k = get_sample_frame_float_string_index(10000)
        pdf_mixed_10k, sff_mixed_10k, npf_mixed_10k = get_sample_frame_mixed_string_index()

        pds_float_h2d_str_index, sfs_float_h2d_str_index = get_series_float_h2d_str_index()
        pds_float_h3d_str_index, sfs_float_h3d_str_index = get_series_float_h3d_str_index()


        for k, v in locals().items():
            if k == 'cls' or k.startswith('__'):
                continue
            cls._store[k] = v

        # additional resources
        label_str = list(''.join(x) for x in it.combinations(string.ascii_lowercase, 4))
        cls._store['label_str'] = label_str

        cls._store['label_tuple2_int_10000'] = [(int(x / 10), x)
                for x in range(10000)]
        cls._store['label_tuple3_int_100000'] = [(int(x / 100), int(x / 10), x)
                for x in range(100000)]
        cls._store['label_tuple4_int_100000'] = [(int(x / 1000), int(x / 100), int(x / 10), x)
                for x in range(100000)]

        label_tuple2_str = []
        label_tuple3_str = []
        label_tuple4_str = []

        for i, label in enumerate(label_str):
            if i % 10 == 0:
                outer10 = label
            if i % 100 == 0:
                outer100 = label
            if i % 1000 == 0:
                outer1000 = label

            label_tuple2_str.append((outer10, label))
            label_tuple3_str.append((outer100, outer10, label))
            label_tuple4_str.append((outer1000, outer100, outer10, label))

        cls._store['label_tuple2_str'] = label_tuple2_str
        cls._store['label_tuple3_str'] = label_tuple3_str
        cls._store['label_tuple4_str'] = label_tuple4_str



    @classmethod
    def get(cls, key: str) -> tp.Any:
        return cls._store[key]



#-------------------------------------------------------------------------------
# index Tests

class IndexStr_init(PerfTest):
    '''Index construction for string labels.
    '''

    @classmethod
    def pd(cls) -> None:
        pd.Index(SampleData.get('label_str'))

    @classmethod
    def sf(cls) -> None:
        sf.Index(SampleData.get('label_str'))


class IndexHierarchy2d_from_product(PerfTest):

    NUMBER = 100

    _size0 = 100
    _size1 = 1000

    @classmethod
    def pd(cls) -> None:
        labels0 = SampleData.get('label_str')[:cls._size0]
        labels1 = SampleData.get('label_str')[:cls._size1]
        ih = pd.MultiIndex.from_product((labels0, labels1))
        assert len(ih) == cls._size0 * cls._size1

    @classmethod
    def sf(cls) -> None:
        labels0 = SampleData.get('label_str')[:cls._size0]
        labels1 = SampleData.get('label_str')[:cls._size1]
        ih = sf.IndexHierarchy.from_product(labels0, labels1)
        assert len(ih) == cls._size0 * cls._size1

class IndexHierarchy3d_from_product(PerfTest):

    NUMBER = 10

    _size0 = 10
    _size1 = 100
    _size2 = 1000

    @classmethod
    def pd(cls) -> None:
        labels0 = SampleData.get('label_str')[:cls._size0]
        labels1 = SampleData.get('label_str')[:cls._size1]
        labels2 = SampleData.get('label_str')[:cls._size2]
        ih = pd.MultiIndex.from_product((labels0, labels1, labels2))
        assert len(ih) == cls._size0 * cls._size1 * cls._size2

    @classmethod
    def sf(cls) -> None:
        labels0 = SampleData.get('label_str')[:cls._size0]
        labels1 = SampleData.get('label_str')[:cls._size1]
        labels2 = SampleData.get('label_str')[:cls._size2]
        ih = sf.IndexHierarchy.from_product(labels0, labels1, labels2)
        assert len(ih) == cls._size0 * cls._size1 * cls._size2


class IndexHierarchy4d_from_product(PerfTest):

    NUMBER = 10

    _size0 = 10
    _size1 = 50
    _size2 = 100
    _size3 = 500

    @classmethod
    def pd(cls) -> None:
        labels0 = SampleData.get('label_str')[:cls._size0]
        labels1 = SampleData.get('label_str')[:cls._size1]
        labels2 = SampleData.get('label_str')[:cls._size2]
        labels3 = SampleData.get('label_str')[:cls._size3]
        ih = pd.MultiIndex.from_product((labels0, labels1, labels2, labels3))
        assert len(ih) == cls._size0 * cls._size1 * cls._size2 * cls._size3

    @classmethod
    def sf(cls) -> None:
        labels0 = SampleData.get('label_str')[:cls._size0]
        labels1 = SampleData.get('label_str')[:cls._size1]
        labels2 = SampleData.get('label_str')[:cls._size2]
        labels3 = SampleData.get('label_str')[:cls._size3]
        ih = sf.IndexHierarchy.from_product(labels0, labels1, labels2, labels3)
        assert len(ih) == cls._size0 * cls._size1 * cls._size2 * cls._size3



class IndexHierarchy2d_from_labels_int(PerfTest):

    NUMBER = 20

    @classmethod
    def pd(cls) -> None:
        ih = pd.MultiIndex.from_tuples(SampleData.get('label_tuple2_int_10000'))

    @classmethod
    def sf(cls) -> None:
        ih = sf.IndexHierarchy.from_labels(SampleData.get('label_tuple2_int_10000'))


class IndexHierarchy2d_from_labels_str(PerfTest):

    NUMBER = 20

    @classmethod
    def pd(cls) -> None:
        ih = pd.MultiIndex.from_tuples(SampleData.get('label_tuple2_str'))

    @classmethod
    def sf(cls) -> None:
        ih = sf.IndexHierarchy.from_labels(SampleData.get('label_tuple2_str'))


class IndexHierarchy3d_from_labels_int(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:
        ih = pd.MultiIndex.from_tuples(SampleData.get('label_tuple3_int_100000'))

    @classmethod
    def sf(cls) -> None:
        ih = sf.IndexHierarchy.from_labels(SampleData.get('label_tuple3_int_100000'))



class IndexHierarchy3d_from_labels_str(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:
        ih = pd.MultiIndex.from_tuples(SampleData.get('label_tuple3_str'))

    @classmethod
    def sf(cls) -> None:
        ih = sf.IndexHierarchy.from_labels(SampleData.get('label_tuple3_str'))


class IndexHierarchy4d_from_labels_int(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:
        ih = pd.MultiIndex.from_tuples(SampleData.get('label_tuple4_int_100000'))

    @classmethod
    def sf(cls) -> None:
        ih = sf.IndexHierarchy.from_labels(SampleData.get('label_tuple4_int_100000'))


class IndexHierarchy4d_from_labels_str(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:
        ih = pd.MultiIndex.from_tuples(SampleData.get('label_tuple4_str'))

    @classmethod
    def sf(cls) -> None:
        ih = sf.IndexHierarchy.from_labels(SampleData.get('label_tuple4_str'))







class IndexHierarchy2d_iter(PerfTest):
    NUMBER = 20

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_float_h2d_str_index').index
        tuple(post)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_float_h2d_str_index').index
        tuple(post)


class IndexHierarchy3d_iter(PerfTest):
    NUMBER = 10

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_float_h3d_str_index').index
        tuple(post)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_float_h3d_str_index').index
        tuple(post)








#-------------------------------------------------------------------------------
# series tests

class SeriesIntFloat_init(PerfTest):
    @staticmethod
    def pd() -> None:
        post = pd.Series(SampleData.get('npa_int_float_10k'))

    @staticmethod
    def sf() -> None:
        post = pd.Series(SampleData.get('npa_int_float_10k'))


class SeriesStrObj_init(PerfTest):
    @staticmethod
    def pd() -> None:
        a = SampleData.get('npa_obj_10k')
        post = pd.Series(a, index=SampleData.get('label_str')[:len(a)])

    @staticmethod
    def sf() -> None:
        a = SampleData.get('npa_obj_10k')
        post = sf.Series(a, index=SampleData.get('label_str')[:len(a)])




class SeriesIntFloat_isnull(PerfTest):
    @staticmethod
    def np() -> None:
        post = np.isnan(SampleData.get('npa_int_float_10k'))

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_int_float_10k').isnull()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_int_float_10k').isna()


class SeriesIntFloat_dropna(PerfTest):
    @staticmethod
    def np() -> None:
        post = SampleData.get('npa_int_float_10k')[np.isnan(SampleData.get('npa_int_float_10k'))]

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_int_float_10k').dropna()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_int_float_10k').dropna()


class SeriesIntFloat_fillna(PerfTest):
    @staticmethod
    def np() -> None:
        sel = np.isnan(SampleData.get('npa_int_float_10k'))
        post = SampleData.get('npa_int_float_10k').copy()
        post[sel] = 0.0

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_int_float_10k').fillna(0.0)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_int_float_10k').fillna(0.0)


class SeriesIntFloat_fillna_forward(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_int_float_10k').fillna(method='ffill')

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_int_float_10k').fillna_forward()


class SeriesIntFloat_drop_duplicated(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_int_float_10k').drop_duplicates(keep=False)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_int_float_10k').drop_duplicated()



class SeriesIntFloat_apply(PerfTest):

    NUMBER = 100

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_int_float_10k').apply(str)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_int_float_10k').iter_element().apply(str)






class SeriesStrFloat_isna(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_str_float_10k').isnull()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_str_float_10k').isna()


class SeriesStrFloat_dropna(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_str_float_10k').dropna()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_str_float_10k').dropna()


class SeriesStrFloat_fillna(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_str_float_10k').fillna(0.0)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_str_float_10k').fillna(0.0)


class SeriesStrFloat_fillna_forward(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_str_float_10k').fillna(method='ffill')

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_str_float_10k').fillna_forward()



class SeriesStrFloat_apply(PerfTest):

    NUMBER = 100

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_str_float_10k').apply(str)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_str_float_10k').iter_element().apply(str)




class SeriesIntObj_isnull(PerfTest):
    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_obj_10k').isnull()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_obj_10k').isna()


class SeriesIntObj_dropna(PerfTest):
    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_obj_10k').dropna()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_obj_10k').dropna()


class SeriesIntObj_fillna(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_obj_10k').fillna(0.0)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_obj_10k').fillna(0.0)


class SeriesIntObj_fillna_forward(PerfTest):
    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_obj_10k').fillna(method='ffill')

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_obj_10k').fillna_forward()


class SeriesIntObj_drop_duplicated(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_obj_10k').drop_duplicates(keep=False)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_obj_10k').drop_duplicated()


class SeriesIntObj_apply(PerfTest):

    NUMBER = 100

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_obj_10k').apply(str)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_obj_10k').iter_element().apply(str)






class SeriesIntObjStr_isnull(PerfTest):
    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_objstr_10k').isnull()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_objstr_10k').isna()

class SeriesIntObjStr_dropna(PerfTest):
    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_objstr_10k').dropna()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_objstr_10k').dropna()

class SeriesIntObjStr_fillna(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_objstr_10k').fillna('wrong')

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_objstr_10k').fillna('wrong')


class SeriesIntObjStr_fillna_forward(PerfTest):
    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_objstr_10k').fillna(method='ffill')

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_objstr_10k').fillna_forward()


class SeriesIntObjStr_apply(PerfTest):

    NUMBER = 100

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_objstr_10k').apply(str)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_objstr_10k').iter_element().apply(str)


class SeriesIntObjStr_reindex_identity(PerfTest):

    NUMBER = 200

    @staticmethod
    def pd() -> None:
        s1 = SampleData.get('pds_objstr_10k')
        s2 = s1.reindex(s1.index)

    @staticmethod
    def sf() -> None:
        s1 = SampleData.get('sfs_objstr_10k')
        s2 = s1.reindex(s1.index)

class SeriesIntObjStr_reindex_reverse(PerfTest):

    NUMBER = 200

    @staticmethod
    def pd() -> None:
        s1 = SampleData.get('pds_objstr_10k')
        s2 = s1.reindex(reversed(s1.index))

    @staticmethod
    def sf() -> None:
        s1 = SampleData.get('sfs_objstr_10k')
        s2 = s1.reindex(reversed(s1.index))



class SeriesFloatH2DString_loc_target(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_float_h2d_str_index').loc[('abgu', 'abcf')]

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_float_h2d_str_index').loc[('abgu', 'abcf')]


class SeriesFloatH2DString_loc_slice(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_float_h2d_str_index').loc[pd.IndexSlice[:, 'abcf']]

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_float_h2d_str_index').loc[sf.HLoc[:, 'abcf']]




class SeriesFloatH3DString_loc_target(PerfTest):
    '''
    Selecting single value from 3-level hierarchy.
    '''
    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_float_h3d_str_index').loc[('abce', 'abgu', 'afgx')]

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_float_h3d_str_index').loc[('abce', 'abgu', 'afgx')]


class SeriesFloatH3DString_loc_slice_target_slice(PerfTest):

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_float_h3d_str_index').loc[
                pd.IndexSlice[:, 'abcf', 'abcl':'abco']]  # type: ignore  # https://github.com/python/typeshed/pull/3024
        assert len(post) == 40

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_float_h3d_str_index').loc[
                sf.HLoc[:, 'abcf', 'abcl':'abco']]  # type: ignore  # https://github.com/python/typeshed/pull/3024
        assert len(post) == 40


class SeriesFloatH3DString_loc_slice_slice_target(PerfTest):
    NUMBER = 50

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pds_float_h3d_str_index').loc[
                pd.IndexSlice[:, :, 'abcl']]
        assert len(post) == 1000

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sfs_float_h3d_str_index').loc[
                sf.HLoc[:, :, 'abcl']]
        assert len(post) == 1000



class SeriesIntObj_isin(PerfTest):
    '''isin with objects.
    Numpy's performance as the lookup array grows deteriorates at an exponential rate.
    '''
    NUMBER = 50
    _lower = 2
    _upper = 5

    @classmethod
    def sf(cls) -> None:
        sf_series = SampleData.get('sfs_obj_10k')
        for i in range(cls._lower, cls._upper):
            lookup_arr = np.array([str(i) for i in range(10**i)], dtype=object)
            sf_series.isin(lookup_arr)

    @classmethod
    def pd(cls) -> None:
        pd_series = SampleData.get('pds_obj_10k')
        for i in range(cls._lower, cls._upper):
            lookup_arr = np.array([str(i) for i in range(10**i)], dtype=object)
            pd_series.isin(lookup_arr)

    # @classmethod
    # def np(cls) -> None:
    #     np_series = SampleData.get('npa_obj_10k')
    #     for i in range(cls._lower, cls._upper):
    #         lookup_arr = np.array([str(i) for i in range(10**i)], dtype=object)
    #         np.isin(np_series, lookup_arr)


class SeriesStrFloat_isin(PerfTest):
    '''isin with primitives.
    Static Frame's performance should be in line with numpy

    As n gets large, static frame begins to outperform pandas
    100 = 1.7124
    1000 = 1.4958
    10000 = 1.1766
    100000 = 1.0496
    1000000 = 0.9719
    '''
    NUMBER = 50
    _lower = 2
    _upper = 5

    @classmethod
    def sf(cls) -> None:
        sf_series = SampleData.get('sfs_str_float_10k')
        for exponent in range(cls._lower, cls._upper):
            lookup_arr = np.array([i / 100 for i in range(10**exponent)])
            sf_series.isin(lookup_arr)

    @classmethod
    def pd(cls) -> None:
        pd_series = SampleData.get('pds_str_float_10k')
        for exponent in range(cls._lower, cls._upper):
            lookup_arr = np.array([i / 100 for i in range(10**exponent)])
            pd_series.isin(lookup_arr)

    @classmethod
    def np(cls) -> None:
        np_series = SampleData.get('npa_str_float_10k')
        for exponent in range(cls._lower, cls._upper):
            lookup_arr = np.array([i / 100 for i in range(10**exponent)])
            np.isin(np_series, lookup_arr)


#-------------------------------------------------------------------------------
# frame tests

# this is deemed invalid, as Pandas just holds a reference
# class FrameFloat_init(PerfTest):
#     @staticmethod
#     def pd() -> None:
#         post = pd.DataFrame(SampleData.get('npf_float_10k'))

#     @staticmethod
#     def sf() -> None:
#         post = sf.Frame(SampleData.get('npf_float_10k'))


class FrameStrFloat_init(PerfTest):
    NUMBER = 100

    @staticmethod
    def pd() -> None:
        data = SampleData.get('npf_float_10k')
        labels = SampleData.get('label_str')
        post = pd.DataFrame(data, index=labels[:data.shape[0]], columns=labels[:data.shape[1]])

    @staticmethod
    def sf() -> None:
        data = SampleData.get('npf_float_10k')
        labels = SampleData.get('label_str')
        post = sf.Frame(data, index=labels[:data.shape[0]], columns=labels[:data.shape[1]])


class FrameFloat_from_records(PerfTest):

    NUMBER = 10

    @staticmethod
    def pd() -> None:
        # make data into a list to force type identification
        post = pd.DataFrame.from_records(list(SampleData.get('npf_float_10k')))
        assert post.shape == (10000, 100)

    @staticmethod
    def sf() -> None:
        post = sf.Frame.from_records(list(SampleData.get('npf_float_10k')), dtypes=[float]*100)
        assert post.shape == (10000, 100)



class FrameMixed_from_records(PerfTest):

    NUMBER = 10

    @staticmethod
    def pd() -> None:
        # make data into a list to force type identification
        post = pd.DataFrame.from_records(list(SampleData.get('npf_mixed_10k')))
        assert post.shape == (10000, 100)
        assert post.dtypes[2] == object

    @staticmethod
    def sf() -> None:
        post = sf.Frame.from_records(list(SampleData.get('npf_mixed_10k')))
        assert post.shape == (10000, 100)
        assert post.dtypes[2] == object





#-------------------------------------------------------------------------------
# frame util functions

class FrameFloat_sum_skipna_axis0(PerfTest):
    @staticmethod
    def np() -> None:
        post = np.nansum(SampleData.get('npf_float_10k'), axis=0)
        assert post.shape == (100,)

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pdf_float_10k').sum(axis=0, skipna=True)
        assert post.shape == (100,)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sff_float_10k').sum(axis=0, skipna=True)
        assert post.shape == (100,)


class FrameFloat_sum_skipna_axis1(PerfTest):
    @staticmethod
    def np() -> None:
        post = np.nansum(SampleData.get('npf_float_10k'), axis=1)
        assert post.shape == (10000,)

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pdf_float_10k').sum(axis=1, skipna=True)
        assert post.shape == (10000,)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sff_float_10k').sum(axis=1, skipna=True)
        assert post.shape == (10000,)


class FrameFloat_dropna_any_axis0(PerfTest):
    NUMBER = 100

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pdf_float_10k').dropna(axis=0, how='any')
        assert post.shape == (9900, 100)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sff_float_10k').dropna(axis=0, condition=np.any)
        assert post.shape == (9900, 100)


class FrameFloat_dropna_any_axis1(PerfTest):
    NUMBER = 100

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pdf_float_10k').dropna(axis=1, how='any')
        assert post.shape == (10000, 50)

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sff_float_10k').dropna(axis=1, condition=np.any)
        assert post.shape == (10000, 50)


class FrameFloat_isna(PerfTest):
    NUMBER = 100

    @staticmethod
    def np() -> None:
        post = np.isnan(SampleData.get('npf_float_10k'))

    @staticmethod
    def pd() -> None:
        post = SampleData.get('pdf_float_10k').isnull()

    @staticmethod
    def sf() -> None:
        post = SampleData.get('sff_float_10k').isna()


class FrameFloat_apply_axis0(PerfTest):

    NUMBER = 50

    @staticmethod
    def pd() -> None:
        func = lambda a: np.nanmean(a ** 2)
        post = SampleData.get('pdf_float_10k').apply(func, axis=0) # apply to columns
        assert post.shape == (100,)
        assert post.sum() > 33501616.16668333

    @staticmethod
    def sf() -> None:
        func = lambda a: np.nanmean(a ** 2)
        post = SampleData.get('sff_float_10k').iter_array(axis=0).apply(func) # apply to columns
        assert post.shape == (100,)
        assert post.sum() > 33501616.16668333


class FrameFloat_apply_axis1(PerfTest):
    NUMBER = 5

    @staticmethod
    def pd() -> None:
        func = lambda a: np.nanmean(a ** 2)
        post = SampleData.get('pdf_float_10k').apply(func, axis=1) # apply to rows
        assert post.shape == (10000,)
        assert post.sum() > 3333328333.8349

    @staticmethod
    def sf() -> None:
        func = lambda a: np.nanmean(a ** 2)
        post = SampleData.get('sff_float_10k').iter_array(axis=1).apply(func)
        assert post.shape == (10000,)
        assert post.sum() > 3333328333.8349


#-------------------------------------------------------------------------------
# frame loc float


class FrameFloat_slice_loc_indices(PerfTest):
    NUMBER = 100

    @staticmethod
    def pd() -> None:
        for i in range(0, 10000, 1000):
            start = SampleData.get('pdf_float_10k').index[i]
            SampleData.get('pdf_float_10k').loc[start:]

    @staticmethod
    def sf() -> None:
        for i in range(0, 10000, 1000):
            start = SampleData.get('sff_float_10k').index.values[i]
            SampleData.get('sff_float_10k').loc[start:]


class FrameFloat_slice_loc_index(PerfTest):
    NUMBER = 100

    @staticmethod
    def pd() -> None:
        for i in range(0, 10000, 1000):
            start = SampleData.get('pdf_float_10k').index[i]
            SampleData.get('pdf_float_10k').loc[start]

    @staticmethod
    def sf() -> None:
        for i in range(0, 10000, 1000):
            start = SampleData.get('sff_float_10k').index.values[i]
            SampleData.get('sff_float_10k').loc[start]



class FrameFloat_slice_loc_columns(PerfTest):

    NUMBER = 100

    @staticmethod
    def pd() -> None:
        for i in range(0, 100, 10):
            start = SampleData.get('pdf_float_10k').index[i]
            SampleData.get('pdf_float_10k').loc[:, start:]

    @staticmethod
    def sf() -> None:
        for i in range(0, 100, 10):
            start = SampleData.get('sff_float_10k').index.values[i]
            SampleData.get('sff_float_10k').loc[:, start:]



class FrameFloat_slice_loc_column(PerfTest):

    @staticmethod
    def pd() -> None:
        for i in range(0, 100, 10):
            start = SampleData.get('pdf_float_10k').index[i]
            SampleData.get('pdf_float_10k').loc[:, start]

    @staticmethod
    def sf() -> None:
        for i in range(0, 100, 10):
            start = SampleData.get('sff_float_10k').index.values[i]
            SampleData.get('sff_float_10k').loc[:, start]


#-------------------------------------------------------------------------------
# frame loc mixed


class FrameMixed_slice_loc_indices(PerfTest):

    NUMBER = 100

    @staticmethod
    def pd() -> None:
        for i in range(0, 10000, 1000):
            start = SampleData.get('pdf_mixed_10k').index[i]
            SampleData.get('pdf_mixed_10k').loc[start:]

    @staticmethod
    def sf() -> None:
        for i in range(0, 10000, 1000):
            start = SampleData.get('sff_mixed_10k').index.values[i]
            SampleData.get('sff_mixed_10k').loc[start:]


class FrameMixed_slice_loc_index(PerfTest):

    NUMBER = 100

    @staticmethod
    def pd() -> None:
        for i in range(0, 10000, 1000):
            start = SampleData.get('pdf_mixed_10k').index[i]
            SampleData.get('pdf_mixed_10k').loc[start]

    @staticmethod
    def sf() -> None:
        for i in range(0, 10000, 1000):
            start = SampleData.get('sff_mixed_10k').index.values[i]
            SampleData.get('sff_mixed_10k').loc[start]



class FrameMixed_slice_loc_columns(PerfTest):

    NUMBER = 100

    @staticmethod
    def pd() -> None:
        for i in range(0, 100, 10):
            start = SampleData.get('pdf_mixed_10k').index[i]
            SampleData.get('pdf_mixed_10k').loc[:, start:]

    @staticmethod
    def sf() -> None:
        for i in range(0, 100, 10):
            start = SampleData.get('sff_mixed_10k').index.values[i]
            SampleData.get('sff_mixed_10k').loc[:, start:]



class FrameMixed_slice_loc_column(PerfTest):

    @staticmethod
    def pd() -> None:
        for i in range(0, 100, 10):
            start = SampleData.get('pdf_mixed_10k').index[i]
            SampleData.get('pdf_mixed_10k').loc[:, start]

    @staticmethod
    def sf() -> None:
        for i in range(0, 100, 10):
            start = SampleData.get('sff_mixed_10k').index.values[i]
            SampleData.get('sff_mixed_10k').loc[:, start]



#-------------------------------------------------------------------------------
# frame creation and growth

class FrameFloat_H1D_add_series_partial(PerfTest):
    '''Adding series that only partially match the index
    '''

    NUMBER = 10

    # 325 two character strings
    _index = list(''.join(x) for x in it.combinations(string.ascii_lowercase, 2))

    @classmethod
    def pd(cls) -> None:
        f1 = pd.DataFrame(index=cls._index)
        for col in range(100):
            s = pd.Series(col * .1, index=cls._index[col: col+20])
            f1[col] = s
        assert f1.sum().sum() == 9900.0

    @classmethod
    def sf(cls) -> None:
        f1 = sf.FrameGO(index=cls._index)
        for col in range(100):
            s = sf.Series.from_element(col * .1, index=cls._index[col: col+20])
            f1[col] = s
        assert f1.sum().sum() == 9900.0


class FrameFloat_H2D_add_series_partial(PerfTest):
    '''Adding series that only partially match the index
    '''
    NUMBER = 10
    _index_leaves = list(''.join(x) for x in it.combinations(string.ascii_lowercase, 2))

    @classmethod
    def sf(cls) -> None:
        index = sf.IndexHierarchy.from_product(list(string.ascii_lowercase),
                list(string.ascii_lowercase))
        f1 = sf.FrameGO(index=index)
        for col in range(100):
            s = sf.Series.from_element(col * .1, index=index[col: col+6]) #pylint: disable=E1136
            f1[col] = s
        assert f1.sum().sum() == 2970.0

    @classmethod
    def pd(cls) -> None:
        index = pd.MultiIndex.from_product((list(string.ascii_lowercase),
                list(string.ascii_lowercase)))
        f1 = pd.DataFrame(index=index)
        for col in range(100):
            s = pd.Series(col * .1, index=index[col: col+6])
            f1[col] = s
        assert f1.sum().sum() == 2970.0


class FrameObj_isin(PerfTest):
    '''isin with objects.
    Will noticeably underperform pandas due to pandas' use of C at a constant rate
    Numpy's performance as the lookup array grows deteriorates at an exponential rate.
    '''
    NUMBER = 5
    _lower = 2
    _upper = 5

    @classmethod
    def pd(cls) -> None:
        pd_frame = SampleData.get('pdf_mixed_10k')
        for i in range(cls._lower, cls._upper):
            lookup_arr = np.array([str(i) for i in range(10 ** i)], dtype=object)
            pd_frame.isin(lookup_arr)

    @classmethod
    def sf(cls) -> None:
        sf_frame = SampleData.get('sff_mixed_10k')
        for i in range(cls._lower, cls._upper):
            lookup_arr = np.array([str(i) for i in range(10 ** i)], dtype=object)
            sf_frame.isin(lookup_arr)

class FrameFloat_isin(PerfTest):
    '''isin with floats. As n gets large, pandas outperformance significantly drops:

    100 = 6.2133x
    1000 = 3.5506x
    10000 = 2.9428x
    100000 = 2.4097x
    1000000 = 1.3527x
    '''

    NUMBER = 5

    _lower = 2
    _upper = 7

    @classmethod
    def pd(cls) -> None:
        pd_frame = SampleData.get('sff_float_10k')
        for i in range(cls._lower, cls._upper):
            lookup_arr = np.array([i / 100 for i in range(10 ** i)])
            pd_frame.isin(lookup_arr)

    @classmethod
    def sf(cls) -> None:
        sf_frame = SampleData.get('sff_float_10k')
        for i in range(cls._lower, cls._upper):
            lookup_arr = np.array([i / 100 for i in range(10 ** i)])
            sf_frame.isin(lookup_arr)

    @classmethod
    def np(cls) -> None:
        np_frame = SampleData.get('npf_float_10k')
        for i in range(cls._lower, cls._upper):
            lookup_arr = np.array([i / 100 for i in range(10 ** i)])
            np.isin(np_frame, lookup_arr)


class FrameFrame_clip(PerfTest):
    '''isin with objects.
    Will noticeably underperform pandas due to pandas' use of C at a constant rate
    Numpy's performance as the lookup array grows deteriorates at an exponential rate.
    '''
    @classmethod
    def pd(cls) -> None:
        pd_frame = SampleData.get('sff_float_10k')
        pd_frame.clip(lower=0, upper=1)

    @classmethod
    def sf(cls) -> None:
        sf_frame = SampleData.get('sff_float_10k')
        sf_frame.clip(lower=0, upper=1)

    @classmethod
    def np(cls) -> None:
        np_frame = SampleData.get('npf_float_10k')
        np.clip(np_frame, 0, 1)





# Always prefer setuptools over distutils
from setuptools import setup
# To use a consistent encoding
from codecs import open
from os import path
import typing as tp

# https://packaging.python.org/distributing/
# to deploy:
# pip install wheel, twine
# python setup.py sdist
# python setup.py bdist_wheel
# twine upload dist/*
# rm -r build; rm -r dist; rm -r *.egg-info

# in /static-frame-feedstock/recipe
# update meta.yaml in feedstock: set version and tar sha256 for tar, commit and push
# submit PR to conda-forge/static-frame-feedstock from fork
# merge into conda forge feedstock after all checks pass

ROOT_DIR_FP = path.abspath(path.dirname(__file__))

def get_long_description() -> str:
    with open(path.join(ROOT_DIR_FP, 'README.rst'), encoding='utf-8') as f:
        msg = []
        collect = False
        start = -1
        for i, line in enumerate(f):
            if line.startswith('static-frame'):
                start = i + 2 # skip this line and the next
            if i == start:
                collect = True
            if line.startswith('Installation'):
                collect = False
            if collect:
                msg.append(line)

    return ''.join(msg).strip()


def get_version() -> str:
    with open(path.join(ROOT_DIR_FP, 'static_frame', '__init__.py'),
            encoding='utf-8') as f:
        for l in f:
            if l.startswith('__version__'):
                if '#' in l:
                    l = l.split('#')[0].strip()
                return l.split('=')[-1].strip()[1:-1]
    raise ValueError("__version__ not found!")


def _get_requirements(file_name: str) -> tp.Iterator[str]:
    with open(path.join(ROOT_DIR_FP, file_name)) as f:
        for line in f:
            line = line.strip()
            if line:
                yield line

def get_install_requires() -> tp.Iterator[str]:
    yield from _get_requirements('requirements.txt')

def get_extras_require() -> tp.Dict[str, tp.List[str]]:
    # For now, have only one group that installs all extras; in the future, can create specialized groups if necessary.
    return {'extras': list(_get_requirements('requirements-extras.txt'))}

setup(
    name='static-frame',
    version=get_version(),
    description='Immutable and grow-only Pandas-like DataFrames with a more explicit and consistent interface',
    long_description=get_long_description(),
    python_requires='>3.6.0',
    install_requires=list(get_install_requires()),
    extras_require=get_extras_require(),
    url='https://github.com/InvestmentSystems/static-frame',
    author='Christopher Ariza',
    license='MIT',
    package_data={'static_frame': ['py.typed']},

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Topic :: Software Development',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Information Analysis',
            'License :: OSI Approved :: MIT License',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            ],

    keywords='staticframe pandas numpy immutable array',
    packages=[
            'static_frame',
            'static_frame.core',
            'static_frame.performance',
            'static_frame.test', # needed for doc generation
            'static_frame.test.unit', # needed for doc generation
            ],
    )



import typing as tp
import warnings

import numpy as np

# from static_frame.core.doc_str import doc_inject
from static_frame.core.frame import Frame
from static_frame.core.store import Store
from static_frame.core.store import store_coherent_non_write
from static_frame.core.store import store_coherent_write
from static_frame.core.store import StoreConfigMap
from static_frame.core.store import StoreConfigMapInitializer
from static_frame.core.type_blocks import TypeBlocks
from static_frame.core.util import DTYPE_STR_KINDS


class StoreHDF5(Store):

    _EXT: tp.FrozenSet[str] =  frozenset(('.h5', '.hdf5'))

    @store_coherent_write
    def write(self,
            items: tp.Iterable[tp.Tuple[tp.Hashable, Frame]],
            *,
            config: StoreConfigMapInitializer = None,
            # store_filter: tp.Optional[StoreFilter] = STORE_FILTER_DEFAULT
            ) -> None:

        config_map = StoreConfigMap.from_initializer(config)

        import tables

        with tables.open_file(self._fp, mode='w') as file, warnings.catch_warnings():
            # silence NaturalNameWarning: object name is not a valid Python identifier:
            warnings.simplefilter("ignore")

            for label, frame in items:
                c = config_map[label]
                label = config_map.default.label_encode(label)

                # should all tables be under a common group?
                field_names, dtypes = self.get_field_names_and_dtypes(
                        frame=frame,
                        include_index=c.include_index,
                        include_index_name=True,
                        include_columns=c.include_columns,
                        include_columns_name=False,
                        )

                # Must set pos to have stable position
                description = {}
                for i, (k, v) in enumerate(zip(field_names, dtypes)):
                    if v == object:
                        raise RuntimeError('cannot store object dtypes in HDF5')
                    description[k] = tables.Col.from_dtype(v, pos=i)

                table = file.create_table('/', # create off root from sring
                        name=label,
                        description=description,
                        expectedrows=len(frame),
                        )

                values = self._get_row_iterator(frame=frame,
                        include_index=c.include_index)
                table.append(tuple(values()))
                table.flush()

    @store_coherent_non_write
    def read_many(self,
            labels: tp.Iterable[tp.Hashable],
            *,
            config: StoreConfigMapInitializer = None,
            container_type: tp.Type[Frame] = Frame,
            ) -> tp.Iterator[Frame]:
        import tables
        config_map = StoreConfigMap.from_initializer(config)

        with tables.open_file(self._fp, mode='r') as file:
            for label in labels:
                c = config_map[label]
                label_encoded = config_map.default.label_encode(label)

                index_depth = c.index_depth
                columns_depth = c.columns_depth
                consolidate_blocks = c.consolidate_blocks
                if c.dtypes:
                    raise NotImplementedError('using config.dtypes on HDF5 not yet supported')

                index_arrays = []
                columns_labels = []

                table = file.get_node(f'/{label_encoded}')
                colnames = table.cols._v_colnames

                def blocks() -> tp.Iterator[np.ndarray]:
                    for col_idx, colname in enumerate(colnames):
                        # can also do: table.read(field=colname)
                        array = table.col(colname)
                        if array.dtype.kind in DTYPE_STR_KINDS:
                            array = array.astype(str)
                        array.flags.writeable = False

                        if col_idx < index_depth:
                            index_arrays.append(array)
                            continue

                        # only store column labels for those yielded
                        columns_labels.append(colname)
                        yield array

                if consolidate_blocks:
                    data = TypeBlocks.from_blocks(TypeBlocks.consolidate_blocks(blocks()))
                else:
                    data = TypeBlocks.from_blocks(blocks())

                # this will own_data in subsequent constructor call
                yield container_type._from_data_index_arrays_column_labels(
                        data=data,
                        index_depth=index_depth,
                        index_arrays=index_arrays,
                        columns_depth=columns_depth,
                        columns_labels=columns_labels,
                        name=label,
                        )

    @store_coherent_non_write
    def labels(self, *,
            config: StoreConfigMapInitializer = None,
            strip_ext: bool = True,
            ) -> tp.Iterator[tp.Hashable]:
        '''
        Iterator of labels.
        '''
        import tables

        config_map = StoreConfigMap.from_initializer(config)

        with tables.open_file(self._fp, mode='r') as file:
            for node in file.iter_nodes(where='/',
                    classname=tables.Table.__name__):
                # NOTE: this is not the complete path
                yield config_map.default.label_decode(node.name)




import typing as tp
from functools import partial
from collections import defaultdict
from itertools import repeat
from itertools import product

import numpy as np
from arraykit import resolve_dtype
from arraykit import resolve_dtype_iter

from static_frame.core.index_base import IndexBase
from static_frame.core.index import Index
from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.util import DepthLevelSpecifier
from static_frame.core.util import IndexConstructor
from static_frame.core.util import UFunc


if tp.TYPE_CHECKING:
    from static_frame.core.frame import Frame #pylint: disable=W0611 #pragma: no cover



#-------------------------------------------------------------------------------
# for Frame.pivot
def extrapolate_column_fields(
        columns_fields: tp.Sequence[tp.Hashable],
        group: tp.Tuple[tp.Hashable, ...],
        data_fields: tp.Sequence[tp.Hashable],
        func_fields: tp.Iterable[tp.Hashable],
        ) -> tp.Iterable[tp.Hashable]:
    '''Used in Frame.pivot.

    Args:
        group: a unique label from the the result of doing a group-by with the `columns_fields`.
    '''
    columns_fields_len = len(columns_fields)
    data_fields_len = len(data_fields)

    sub_columns: tp.Iterable[tp.Hashable]

    if columns_fields_len == 1 and data_fields_len == 1:
        if not func_fields:
            sub_columns = group # already a tuple
        else:
            sub_columns = [group + (label,) for label in func_fields]
    elif columns_fields_len == 1 and data_fields_len > 1: # create a sub heading for each data field
        if not func_fields:
            sub_columns = list(product(group, data_fields))
        else:
            sub_columns = list(product(group, data_fields, func_fields))
    elif columns_fields_len > 1 and data_fields_len == 1:
        if not func_fields:
            sub_columns = (group,)
        else:
            sub_columns = [group + (label,) for label in func_fields]
    else: # group is already a tuple of the partial column label; need to extend with each data field
        if not func_fields:
            sub_columns = [group + (field,) for field in data_fields]
        else:
            sub_columns = [group + (field, label) for field in data_fields for label in func_fields]

    return sub_columns


def pivot_records_dtypes(
        frame: 'Frame',
        data_fields: tp.Iterable[tp.Hashable],
        func_single: tp.Optional[UFunc],
        func_map: tp.Sequence[tp.Tuple[tp.Hashable, UFunc]]
        ) -> tp.Iterator[np.dtype]:
    dtypes = frame.dtypes
    for field in data_fields:
        dtype = dtypes[field]
        if func_single:
            yield dtype
        else: # we assume
            for _, func in func_map:
                yield None # do not know what func result will be


def pivot_records_items(
        frame: 'Frame',
        group_fields: tp.Iterable[tp.Hashable],
        group_depth: int,
        data_fields: tp.Iterable[tp.Hashable],
        func_single: tp.Optional[UFunc],
        func_map: tp.Sequence[tp.Tuple[tp.Hashable, UFunc]]
        ) -> tp.Iterator[tp.Tuple[tp.Hashable, tp.Sequence[tp.Any]]]:

    take_group_index = group_depth > 1

    for group_index, part in frame.iter_group_items(group_fields):
        label = group_index if take_group_index else group_index[0]
        record = []
        part_columns_loc_to_iloc = part.columns._loc_to_iloc
        for field in data_fields:
            values = part._blocks._extract_array(
                    row_key=None,
                    column_key=part_columns_loc_to_iloc(field),
                    )
            if func_single:
                if len(values) == 1:
                    record.append(values[0])
                else:
                    record.append(func_single(values))
            else:
                for _, func in func_map:
                    if len(values) == 1:
                        record.append(values[0])
                    else:
                        record.append(func(values))
        yield label, record

def pivot_items(
        frame: 'Frame',
        group_fields: tp.Iterable[tp.Hashable],
        group_depth: int,
        data_fields: tp.Sequence[tp.Hashable],
        func_single: UFunc,
        ) -> tp.Iterator[tp.Tuple[tp.Hashable, tp.Any]]:
    '''
    Specialized generator of Pairs for when group_fields has been reduced to a single column.
    '''
    take_group = group_depth > 1

    for group, sub in frame.iter_group_items(group_fields):
        label = group if take_group else group[0]
        values = sub._blocks._extract_array(
                row_key=None,
                column_key=sub.columns._loc_to_iloc(data_fields[0]),
                )
        if len(values) == 1:
            yield label, values[0]
        else: # can be sure we only have func_single
            yield label, func_single(values)

#-------------------------------------------------------------------------------
class PivotIndexMap(tp.NamedTuple):
    targets_unique: tp.Iterable[tp.Hashable]
    target_depth: int
    target_select: np.ndarray
    group_to_target_map: tp.Dict[tp.Optional[tp.Hashable], tp.Dict[tp.Any, int]]
    group_depth: int
    group_select: np.ndarray
    group_to_dtype: tp.Dict[tp.Optional[tp.Hashable], np.dtype]

def pivot_index_map(*,
        index_src: IndexBase,
        depth_level: DepthLevelSpecifier,
        dtypes_src: tp.Optional[tp.Sequence[np.dtype]],
        ) -> PivotIndexMap:
    '''
    Args:
        dtypes_src: must be of length equal to axis
    '''
    # We are always moving levels from one axis to another; after application, the expanded axis will always be hierarchical, while the contracted axis may or may not be. From the contract axis, we need to divide the depths into two categories: targets (the depths to be moved and added to expand axis) and groups (unique combinations that remain on the contract axis after removing targets).

    # Unique target labels are added to labels on the expand axis; unique group labels become the new contract axis.

    target_select = np.full(index_src.depth, False)
    target_select[depth_level] = True
    group_select = ~target_select

    group_arrays = []
    target_arrays = []
    for i, v in enumerate(target_select):
        if v:
            target_arrays.append(index_src.values_at_depth(i))
        else:
            group_arrays.append(index_src.values_at_depth(i))

    group_depth = len(group_arrays)
    target_depth = len(target_arrays)
    group_to_dtype: tp.Dict[tp.Optional[tp.Hashable], np.dtype] = {}
    targets_unique: tp.Iterable[tp.Hashable]

    if group_depth == 0:
        # targets must be a tuple
        group_to_target_map = {
                None: {v: idx for idx, v in enumerate(zip(*target_arrays))}
                }
        targets_unique = [k for k in group_to_target_map[None]]
        if dtypes_src is not None:
            group_to_dtype[None] = resolve_dtype_iter(dtypes_src)
    else:
        group_to_target_map = defaultdict(dict)
        targets_unique = dict() # Store targets in order observed

        for axis_idx, (group, target, dtype) in enumerate(zip(
                zip(*group_arrays), # get tuples of len 1 to depth
                zip(*target_arrays),
                (dtypes_src if dtypes_src is not None else repeat(None)),
                )):
            if group_depth == 1:
                group = group[0]
            # targets are transfered labels; groups are the new columns
            group_to_target_map[group][target] = axis_idx
            targets_unique[target] = None #type: ignore

            if dtypes_src is not None:
                if group in group_to_dtype:
                    group_to_dtype[group] = resolve_dtype(group_to_dtype[group], dtype)
                else:
                    group_to_dtype[group] = dtype

    return PivotIndexMap( #pylint: disable=E1120
            targets_unique=targets_unique,
            target_depth=target_depth,
            target_select=target_select,
            group_to_target_map=group_to_target_map, #type: ignore
            group_depth=group_depth,
            group_select=group_select,
            group_to_dtype=group_to_dtype
            )


#-------------------------------------------------------------------------------
class PivotDeriveConstructors(tp.NamedTuple):
    contract_dst: tp.Optional[tp.Iterable[tp.Hashable]]
    contract_constructor: IndexConstructor
    expand_constructor: IndexConstructor

def pivot_derive_constructors(*,
        contract_src: IndexBase,
        expand_src: IndexBase,
        group_select: np.ndarray, # Boolean
        group_depth: int,
        target_select: np.ndarray,
        # target_depth: int,
        group_to_target_map: tp.Dict[tp.Hashable, tp.Tuple[tp.Hashable]],
        expand_is_columns: bool,
        frame_cls: tp.Type['Frame'],
        ) -> PivotDeriveConstructors:
    '''
    pivot_stack: columns is contract, index is expand
    pivot_unstack: index is contract, columns is expand
    '''
    # NOTE: group_select, target_select operate on the contract axis
    if expand_is_columns:
        contract_cls = Index
        contract_cls_hierarchy = IndexHierarchy
        expand_cls_hierarchy = frame_cls._COLUMNS_HIERARCHY_CONSTRUCTOR
    else: # contract is columns
        contract_cls = frame_cls._COLUMNS_CONSTRUCTOR
        contract_cls_hierarchy = frame_cls._COLUMNS_HIERARCHY_CONSTRUCTOR
        expand_cls_hierarchy = IndexHierarchy

    # NOTE: not propagating name attr, as not obvious how it should when depths are exiting and entering

    # contract axis may or may not be IndexHierarchy after extracting depths
    if contract_src.depth == 1: # will removed that one level, thus need IndexAuto
        contract_dst = None
        contract_constructor = contract_cls
    else:
        contract_src_types = contract_src.index_types.values #type: ignore
        contract_dst_types = contract_src_types[group_select]
        if group_depth == 0:
            contract_dst = None
            contract_constructor = contract_cls
        elif group_depth == 1:
            contract_dst = list(group_to_target_map.keys())
            contract_constructor = contract_dst_types[0]
        else:
            contract_dst = list(group_to_target_map.keys())
            contract_constructor = partial( #type: ignore
                    contract_cls_hierarchy.from_labels,
                    index_constructors=contract_dst_types,
                    )

    # expand axis will always be IndexHierarchy after adding depth
    if expand_src.depth == 1:
        expand_types = [expand_src.__class__]
    else:
        expand_types = list(expand_src._levels.index_types()) #type: ignore

    if contract_src.depth == 1:
        expand_types.append(contract_src.__class__)
    else:
        expand_types.extend(contract_src_types[target_select])

    expand_constructor = partial(
            expand_cls_hierarchy.from_labels,
            index_constructors=expand_types,
            # name=expand_src.name,
            )

    # NOTE: expand_dst labels will come from the values generator
    return PivotDeriveConstructors( #pylint: disable=E1120
            contract_dst=contract_dst,
            contract_constructor=contract_constructor,
            expand_constructor=expand_constructor,
            )





import typing as tp

from static_frame.core.doc_str import doc_inject

if tp.TYPE_CHECKING:
    from static_frame.core.frame import Frame #pylint: disable=W0611 #pragma: no cover

class InterfaceMeta(type):
    '''Lowest level metaclass for providing interface property on class.
    '''

    @property #type: ignore
    @doc_inject()
    def interface(cls) -> 'Frame':
        '''{}'''
        from static_frame.core.interface import InterfaceSummary
        return InterfaceSummary.to_frame(cls) #type: ignore



'''
Tools for iterators in Series and Frame. These components are imported by both series.py and frame.py; these components also need to be able to return Series and Frame, and thus use deferred, function-based imports.
'''

import typing as tp
from enum import Enum
from functools import partial
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from arraykit import name_filter

from static_frame.core.doc_str import doc_inject
from static_frame.core.util import AnyCallable
from static_frame.core.util import DepthLevelSpecifier
from static_frame.core.util import DtypeSpecifier
from static_frame.core.util import KEY_ITERABLE_TYPES
from static_frame.core.util import Mapping
from static_frame.core.util import NameType
from static_frame.core.util import TupleConstructorType
from static_frame.core.util import iterable_to_array_1d
# from static_frame.core.util import array_from_iterator


if tp.TYPE_CHECKING:
    from static_frame.core.frame import Frame # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.series import Series # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.quilt import Quilt # pylint: disable=W0611 #pragma: no cover
    from static_frame.core.bus import Bus # pylint: disable=W0611 #pragma: no cover


FrameOrSeries = tp.TypeVar('FrameOrSeries', 'Frame', 'Series', 'Bus', 'Quilt')
PoolArgGen = tp.Callable[[], tp.Union[tp.Iterator[tp.Any], tp.Iterator[tp.Tuple[tp.Any, tp.Any]]]]
# FrameSeriesIndex = tp.TypeVar('FrameSeriesIndex', 'Frame', 'Series', 'Index')


class IterNodeApplyType(Enum):
    SERIES_VALUES = 0
    SERIES_ITEMS = 1
    SERIES_ITEMS_GROUP_VALUES = 2
    SERIES_ITEMS_GROUP_LABELS = 3
    FRAME_ELEMENTS = 4
    INDEX_LABELS = 5

class IterNodeType(Enum):
    VALUES = 1
    ITEMS = 2

class IterNodeDelegate(tp.Generic[FrameOrSeries]):
    '''
    Delegate returned from :obj:`static_frame.IterNode`, providing iteration as well as a family of apply methods.
    '''

    __slots__ = (
            '_func_values',
            '_func_items',
            '_yield_type',
            '_apply_constructor',
            '_apply_type',
            )

    INTERFACE = (
            'apply',
            'apply_iter',
            'apply_iter_items',
            'apply_pool',
            'map_all',
            'map_all_iter',
            'map_all_iter_items',
            'map_any',
            'map_any_iter',
            'map_any_iter_items',
            'map_fill',
            'map_fill_iter',
            'map_fill_iter_items',
            )

    def __init__(self,
            func_values: tp.Callable[..., tp.Iterable[tp.Any]],
            func_items: tp.Callable[..., tp.Iterable[tp.Tuple[tp.Any, tp.Any]]],
            yield_type: IterNodeType,
            apply_constructor: tp.Callable[..., FrameOrSeries],
            apply_type: IterNodeApplyType,
        ) -> None:
        '''
        Args:
            apply_constructor: Callable (generally a class) used to construct the object returned from apply(); must take an iterator of items.
        '''
        self._func_values = func_values
        self._func_items = func_items
        self._yield_type = yield_type
        self._apply_constructor: tp.Callable[..., FrameOrSeries] = apply_constructor
        self._apply_type = apply_type

    #---------------------------------------------------------------------------

    def _apply_iter_items_parallel(self,
            func: AnyCallable,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:

        pool_executor = ThreadPoolExecutor if use_threads else ProcessPoolExecutor

        if not callable(func): # NOTE: when is func not a callable?
            func = getattr(func, '__getitem__')

        # use side effect list population to create keys when iterating over values
        func_keys = []
        arg_gen: PoolArgGen

        if self._yield_type is IterNodeType.VALUES:
            def arg_gen() -> tp.Iterator[tp.Any]: #pylint: disable=E0102
                for k, v in self._func_items():
                    func_keys.append(k)
                    yield v
        else:
            def arg_gen() -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]: #pylint: disable=E0102
                for k, v in self._func_items():
                    func_keys.append(k)
                    yield k, v

        with pool_executor(max_workers=max_workers) as executor:
            yield from zip(func_keys,
                    executor.map(func, arg_gen(), chunksize=chunksize)
                    )

    def _apply_iter_parallel(self,
            func: AnyCallable,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False,
            ) -> tp.Iterator[tp.Any]:

        pool_executor = ThreadPoolExecutor if use_threads else ProcessPoolExecutor

        if not callable(func): # NOTE: when is func not a callable?
            func = getattr(func, '__getitem__')

        # use side effect list population to create keys when iterating over values
        arg_gen = (self._func_values if self._yield_type is IterNodeType.VALUES
                else self._func_items)

        with pool_executor(max_workers=max_workers) as executor:
            yield from executor.map(func, arg_gen(), chunksize=chunksize)

    #---------------------------------------------------------------------------
    # public interface

    @doc_inject(selector='map_any')
    def map_any_iter_items(self,
            mapping: Mapping
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
        '''
        {doc} A generator of resulting key, value pairs.

        Args:
            {mapping}
        '''
        get = getattr(mapping, 'get')
        if self._yield_type is IterNodeType.VALUES:
            yield from ((k, get(v, v)) for k, v in self._func_items())
        else:
            yield from ((k, get((k,  v), v)) for k, v in self._func_items())

    @doc_inject(selector='map_any')
    def map_any_iter(self,
            mapping: Mapping,
            ) -> tp.Iterator[tp.Any]:
        '''
        {doc} A generator of resulting values.

        Args:
            {mapping}
        '''
        get = getattr(mapping, 'get')
        if self._yield_type is IterNodeType.VALUES:
            yield from (get(v, v) for v in self._func_values())
        else:
            yield from (get((k,  v), v) for k, v in self._func_items())

    @doc_inject(selector='map_any')
    def map_any(self,
            mapping: Mapping,
            *,
            dtype: DtypeSpecifier = None, # can be DtypesSpecifier in some contexts
            name: NameType = None,
            ) -> FrameOrSeries:
        '''
        {doc} Returns a new container.

        Args:
            {mapping}
            {dtype}
        '''
        if (self._apply_type is IterNodeApplyType.SERIES_VALUES
                or self._apply_type is IterNodeApplyType.INDEX_LABELS):
            return self._apply_constructor(
                    self.map_any_iter(mapping),
                    dtype=dtype,
                    name=name,
                    )

        return self._apply_constructor(
                self.map_any_iter_items(mapping),
                dtype=dtype,
                name=name,
                )


    #---------------------------------------------------------------------------
    @doc_inject(selector='map_fill')
    def map_fill_iter_items(self,
            mapping: Mapping,
            *,
            fill_value: tp.Any = np.nan,
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
        '''
        {doc} A generator of resulting key, value pairs.

        Args:
            {mapping}
            {fill_value}
        '''
        get = getattr(mapping, 'get')
        if self._yield_type is IterNodeType.VALUES:
            yield from ((k, get(v, fill_value)) for k, v in self._func_items())
        else:
            yield from ((k, get((k,  v), fill_value)) for k, v in self._func_items())

    @doc_inject(selector='map_fill')
    def map_fill_iter(self,
            mapping: Mapping,
            *,
            fill_value: tp.Any = np.nan,
            ) -> tp.Iterator[tp.Any]:
        '''
        {doc} A generator of resulting values.

        Args:
            {mapping}
            {fill_value}
        '''
        get = getattr(mapping, 'get')
        if self._yield_type is IterNodeType.VALUES:
            yield from (get(v, fill_value) for v in self._func_values())
        else:
            yield from (get((k,  v), fill_value) for k, v in self._func_items())


    @doc_inject(selector='map_fill')
    def map_fill(self,
            mapping: Mapping,
            *,
            fill_value: tp.Any = np.nan,
            dtype: DtypeSpecifier = None,  # can be DtypesSpecifier in some contexts
            name: NameType = None,
            ) -> FrameOrSeries:
        '''
        {doc} Returns a new container.

        Args:
            {mapping}
            {fill_value}
            {dtype}
        '''
        if (self._apply_type is IterNodeApplyType.SERIES_VALUES
                or self._apply_type is IterNodeApplyType.INDEX_LABELS):
            return self._apply_constructor(
                    self.map_fill_iter(mapping, fill_value=fill_value),
                    dtype=dtype,
                    name=name,
                    )

        return self._apply_constructor(
                self.map_fill_iter_items(mapping, fill_value=fill_value),
                dtype=dtype,
                name=name,
                )

    #---------------------------------------------------------------------------
    @doc_inject(selector='map_all')
    def map_all_iter_items(self,
            mapping: Mapping
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
        '''
        {doc} A generator of resulting key, value pairs.

        Args:
            {mapping}
        '''
        # want exception to raise if key not found
        func = getattr(mapping, '__getitem__')
        if self._yield_type is IterNodeType.VALUES:
            yield from ((k, func(v)) for k, v in self._func_items())
        else:
            yield from ((k, func((k,  v))) for k, v in self._func_items())

    @doc_inject(selector='map_all')
    def map_all_iter(self,
            mapping: Mapping
            ) -> tp.Iterator[tp.Any]:
        '''
        {doc} A generator of resulting values.

        Args:
            {mapping}
        '''
        func = getattr(mapping, '__getitem__')
        if self._yield_type is IterNodeType.VALUES:
            yield from (func(v) for v in self._func_values())
        else:
            yield from (func((k,  v)) for k, v in self._func_items())

    @doc_inject(selector='map_all')
    def map_all(self,
            mapping: Mapping,
            *,
            dtype: DtypeSpecifier = None,  # can be DtypesSpecifier in some contexts
            name: NameType = None,
            ) -> FrameOrSeries:
        '''
        {doc} Returns a new container.

        Args:
            {mapping}
            {dtype}
        '''
        if (self._apply_type is IterNodeApplyType.SERIES_VALUES
                or self._apply_type is IterNodeApplyType.INDEX_LABELS):
            return self._apply_constructor(
                    self.map_all_iter(mapping),
                    dtype=dtype,
                    name=name,
                    )

        return self._apply_constructor(
                self.map_all_iter_items(mapping),
                dtype=dtype,
                name=name,
                )

    #---------------------------------------------------------------------------
    @doc_inject(selector='apply')
    def apply_iter_items(self,
            func: AnyCallable,
            ) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
        '''
        {doc} A generator of resulting key, value pairs.

        Args:
            {func}
        '''
        # depend on yield type, we determine what the passed in function expects to take
        if self._yield_type is IterNodeType.VALUES:
            yield from ((k, func(v)) for k, v in self._func_items())
        else:
            yield from ((k, func(k, v)) for k, v in self._func_items())

    @doc_inject(selector='apply')
    def apply_iter(self,
            func: AnyCallable
            ) -> tp.Iterator[tp.Any]:
        '''
        {doc} A generator of resulting values.

        Args:
            {func}
        '''
        if self._yield_type is IterNodeType.VALUES:
            yield from (func(v) for v in self._func_values())
        else:
            yield from (func(k, v) for k, v in self._func_items())

    @doc_inject(selector='apply')
    def apply(self,
            func: AnyCallable,
            *,
            dtype: DtypeSpecifier = None,  # can be DtypesSpecifier in some contexts
            name: NameType = None,
            ) -> FrameOrSeries:
        '''
        {doc} Returns a new container.

        Args:
            {func}
            {dtype}
        '''
        if not callable(func):
            raise RuntimeError('use map_fill(), map_any(), or map_all() for applying a mapping type')

        if (self._apply_type is IterNodeApplyType.SERIES_VALUES
                or self._apply_type is IterNodeApplyType.INDEX_LABELS):
            return self._apply_constructor(
                    self.apply_iter(func),
                    dtype=dtype,
                    name=name,
                    )

        # only use when we need pairs of values to dynamically create an Index
        return self._apply_constructor(
                self.apply_iter_items(func),
                dtype=dtype,
                name=name,
                )

    @doc_inject(selector='apply')
    def apply_pool(self,
            func: AnyCallable,
            *,
            dtype: DtypeSpecifier = None,
            name: NameType = None,
            max_workers: tp.Optional[int] = None,
            chunksize: int = 1,
            use_threads: bool = False
            ) -> FrameOrSeries:
        '''
        {doc} Employ parallel processing with either the ProcessPoolExecutor or ThreadPoolExecutor.

        Args:
            {func}
            *
            {dtype}
            {name}
            {max_workers}
            {chunksize}
            {use_threads}
        '''
        if (self._apply_type is IterNodeApplyType.SERIES_VALUES
                or self._apply_type is IterNodeApplyType.INDEX_LABELS):
            return self._apply_constructor(
                    self._apply_iter_parallel(
                            func=func,
                            max_workers=max_workers,
                            chunksize=chunksize,
                            use_threads=use_threads),
                    dtype=dtype,
                    name=name,
                    )

        return self._apply_constructor(
                self._apply_iter_items_parallel(
                        func=func,
                        max_workers=max_workers,
                        chunksize=chunksize,
                        use_threads=use_threads),
                dtype=dtype,
                name=name,
                )

    def __iter__(self) -> tp.Union[
            tp.Iterator[tp.Any],
            tp.Iterator[tp.Tuple[tp.Any, tp.Any]]
            ]:
        '''
        Return a generator based on the yield type.
        '''
        if self._yield_type is IterNodeType.VALUES:
            yield from self._func_values()
        else:
            yield from self._func_items()



#-------------------------------------------------------------------------------

_ITER_NODE_SLOTS = (
        '_container',
        '_func_values',
        '_func_items',
        '_yield_type',
        '_apply_type'
        )

class IterNode(tp.Generic[FrameOrSeries]):
    '''Interface to a type of iteration on :obj:`static_frame.Series` and :obj:`static_frame.Frame`.
    '''
    # Stores two version of a generator function: one to yield single values, another to yield items pairs. The latter is needed in all cases, as when we use apply we return a Series, and need to have recourse to an index.

    __slots__ = _ITER_NODE_SLOTS

    def __init__(self, *,
            container: FrameOrSeries,
            function_values: tp.Callable[..., tp.Iterable[tp.Any]],
            function_items: tp.Callable[..., tp.Iterable[tp.Tuple[tp.Any, tp.Any]]],
            yield_type: IterNodeType,
            apply_type: IterNodeApplyType,
            ) -> None:
        '''
        Args:
            function_values: will be partialed with arguments given with __call__.
            function_items: will be partialed with arguments given with __call__.
        '''
        self._container: FrameOrSeries = container
        self._func_values = function_values
        self._func_items = function_items
        self._yield_type = yield_type
        self._apply_type = apply_type

    def get_delegate(self,
            **kwargs: object
            ) -> IterNodeDelegate[FrameOrSeries]:
        '''
        In usage as an iteator, the args passed here are expected to be argument for the core iterators, i.e., axis arguments.

        Args:
            kwargs: kwarg args to be passed to both self._func_values and self._func_items
        '''
        from static_frame.core.series import Series
        from static_frame.core.frame import Frame
        from static_frame.core.index import Index

        func_values = partial(self._func_values, **kwargs)
        func_items = partial(self._func_items, **kwargs)
        # only some apply_types can use
        shape: tp.Optional[tp.Tuple[int, ...]] = None

        apply_constructor: tp.Callable[..., tp.Union[Frame, Series]]

        if self._apply_type is IterNodeApplyType.SERIES_VALUES:
            # Creating a Series that will have the same index as source container
            if isinstance(self._container, Frame) and kwargs['axis'] == 0:
                index = self._container._columns
                own_index = False
            else:
                index = self._container._index
                own_index = True

            shape = index.shape

            def apply_constructor( #pylint: disable=E0102
                    values: tp.Iterator[tp.Any],
                    dtype: DtypeSpecifier,
                    name: NameType = None,
                    ) -> Series:
                # PERF: passing count here permits faster generator realization
                values, _ = iterable_to_array_1d(
                        values,
                        count=shape[0], # type: ignore
                        dtype=dtype,
                        )
                return Series(values, name=name, index=index, own_index=own_index)

        elif self._apply_type is IterNodeApplyType.SERIES_ITEMS:
            # Only use this path if the Index to be returned is different than the source container
            name_index = None # NOTE: what should this be?
            if isinstance(self._container, Frame) and kwargs['axis'] == 0:
                index_constructor = partial(
                        self._container._columns.from_labels,
                        name=name_index)
            else:
                index_constructor = partial(
                        self._container._index.from_labels,
                        name=name_index)
            # always return a Series
            apply_constructor = partial(
                    Series.from_items,
                    index_constructor=index_constructor
                    )

        elif self._apply_type is IterNodeApplyType.SERIES_ITEMS_GROUP_VALUES:
            # use default index constructor
            # NOTE: when used on labels, this key is given; when used on lables (indices) depth_level is given; only take the key if it is a hashable (a string or a tuple, not a slice, list, or array)
            try:
                name_index = name_filter(kwargs.get('key', None))
            except TypeError:
                name_index = None

            index_constructor = partial(
                    Index.from_labels,
                    name=name_index)
            apply_constructor = partial(
                    Series.from_items,
                    index_constructor=index_constructor,
                    )

        elif self._apply_type is IterNodeApplyType.SERIES_ITEMS_GROUP_LABELS:
            # use default index constructor
            # will always have `depth_level` in kwargs, and for Frame an axis; could attempt to get name from the index if it has a name
            name_index = None

            index_constructor = partial(
                    Index.from_labels,
                    name=name_index)
            apply_constructor = partial(
                    Series.from_items,
                    index_constructor=index_constructor,
                    )

        elif self._apply_type is IterNodeApplyType.FRAME_ELEMENTS:
            assert isinstance(self._container, Frame) # for typing
            # for element-wise function application, axis will always be 0 or 1, as we always do full iteration; from_element_items accepts axis of None for incomplete specification, but that will never be used here.
            apply_constructor = partial(
                    self._container.__class__.from_element_items,
                    index=self._container._index,
                    columns=self._container._columns,
                    axis=kwargs['axis'],
                    own_index=True,
                    index_constructor=self._container._index.from_labels,
                    columns_constructor=self._container._columns.from_labels
                    )
            shape = self._container.shape

        elif self._apply_type is IterNodeApplyType.INDEX_LABELS:
            shape = self._container.shape

            def apply_constructor( #pylint: disable=E0102
                    values: tp.Iterator[tp.Hashable], #pylint: disable=function-redefined
                    dtype: DtypeSpecifier = None,
                    name: NameType = None,
                    ) -> np.ndarray:
                # NOTE: name argument is for common interface
                # PERF: passing count here permits faster generator realization
                array, _ = iterable_to_array_1d(values, count=shape[0], dtype=dtype) #type: ignore
                return array
        else:
            raise NotImplementedError(self._apply_type) #pragma: no cover

        return IterNodeDelegate(
                func_values=func_values,
                func_items=func_items,
                yield_type=self._yield_type,
                apply_constructor=tp.cast(tp.Callable[..., FrameOrSeries], apply_constructor),
                apply_type=self._apply_type,
                )


#-------------------------------------------------------------------------------
# specialize IterNode based on arguments given to __call__

class IterNodeNoArg(IterNode[FrameOrSeries]):

    __slots__ = _ITER_NODE_SLOTS

    def __call__(self,
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self)


class IterNodeAxis(IterNode[FrameOrSeries]):

    __slots__ = _ITER_NODE_SLOTS

    def __call__(self,
            *,
            axis: int = 0
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self, axis=axis)


class IterNodeConstructorAxis(IterNode[FrameOrSeries]):

    __slots__ = _ITER_NODE_SLOTS

    def __call__(self,
            *,
            axis: int = 0,
            constructor: tp.Optional[TupleConstructorType] = None,
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self,
                axis=axis,
                constructor=constructor,
                )

class IterNodeGroup(IterNode[FrameOrSeries]):
    '''
    Iterator on 1D groupings where no args are required (but axis is retained for compatibility)
    '''

    __slots__ = _ITER_NODE_SLOTS

    def __call__(self,
            *,
            axis: int = 0
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self, axis=axis)

class IterNodeGroupAxis(IterNode[FrameOrSeries]):
    '''
    Iterator on 2D groupings where key and axis are required.
    '''

    __slots__ = _ITER_NODE_SLOTS

    def __call__(self,
            key: KEY_ITERABLE_TYPES, # type: ignore
            *,
            axis: int = 0
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self, key=key, axis=axis)


class IterNodeDepthLevel(IterNode[FrameOrSeries]):

    __slots__ = _ITER_NODE_SLOTS

    def __call__(self,
            depth_level: tp.Optional[DepthLevelSpecifier] = None
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self, depth_level=depth_level)


class IterNodeDepthLevelAxis(IterNode[FrameOrSeries]):

    __slots__ = _ITER_NODE_SLOTS

    def __call__(self,
            depth_level: DepthLevelSpecifier = 0,
            *,
            axis: int = 0
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self, depth_level=depth_level, axis=axis)


class IterNodeWindow(IterNode[FrameOrSeries]):

    __slots__ = _ITER_NODE_SLOTS

    def __call__(self, *,
            size: int,
            axis: int = 0,
            step: int = 1,
            window_sized: bool = True,
            window_func: tp.Optional[AnyCallable] = None,
            window_valid: tp.Optional[AnyCallable] = None,
            label_shift: int = 0,
            start_shift: int = 0,
            size_increment: int = 0,
            ) -> IterNodeDelegate[FrameOrSeries]:
        return IterNode.get_delegate(self,
                axis=axis,
                size=size,
                step=step,
                window_sized=window_sized,
                window_func=window_func,
                window_valid=window_valid,
                label_shift=label_shift,
                start_shift=start_shift,
                size_increment=size_increment,
                )

